"""Abstract interfaces for protocol file running."""
from abc import ABC, abstractmethod


class AbstractFileRunner(ABC):
    """Abstract interface for an object that can run protocol files."""

    @abstractmethod
    def load(self) -> None:
        """Prepare runner and engine state prior to starting the run."""
        ...

    @abstractmethod
    def play(self) -> None:
        """Start (or un-pause) running the protocol file."""
        ...

    @abstractmethod
    def pause(self) -> None:
        """Pause the running protocol file's execution."""
        ...

    @abstractmethod
    def stop(self) -> None:
        """Cancel the running protocol file."""
        ...



"""
Z-Stage Subassembly Test:
Travel 3x full z-stage motions @ max machine speed using 80% current.
Record for any lost steps
Travel 3x full z-stage motions @ min pickup tip current *80% current
    @ pickup tip speed. Record for any lost steps


Author: Carlos Fernandez
Date: 3/21/2018
QC Gantry Test

"""

import atexit
import functools
import logging

from opentrons.drivers.smoothie_drivers.driver_3_0 import SmoothieError
from . import args_handler


def setup(motor_current, max_speed, driver):
    driver.set_active_current({"Z": motor_current, "A": motor_current})
    driver.set_axis_max_speed({'Z': max_speed, 'A': max_speed})
    driver.set_speed(max_speed)


def pick_up_motion(max_dist, max_speed, low_speed, driver, options):
    zero = 100
    zero_1 = 5

    # Descent Z z to 100mm
    setup(options.high_current, max_speed, driver)
    driver.set_speed(max_speed)
    driver.move({'Z': zero, 'A': zero})
    # Press Action
    setup(options.low_current, max_speed, driver)
    driver.set_speed(low_speed)
    driver.move({'Z': zero_1, 'A': zero_1})
    # Retract Action
    setup(options.high_current, max_speed, driver)
    driver.set_speed(max_speed)
    driver.move({'Z': zero, 'A': zero})
    # Jog up
    driver.set_speed(max_speed)
    driver.move({'Z': max_dist, 'A': max_dist})


def test_axis(axis, tolerance, driver):
    retract_amounts = {
        'X': 3,
        'Y': 3,
        'Z': 2,
        'A': 2,
        'B': 2,
        'C': 2
    }
    retract = retract_amounts[axis]
    expected_point = driver.homed_position[axis] + retract
    points = [expected_point - tolerance, expected_point + tolerance]
    driver.push_speed()
    driver.set_speed(8)
    try:
        driver.move({axis: points[0]})
    except SmoothieError:
        raise Exception('Test Failed: Pressing too soon')
    if axis == 'Y':
        if driver.switch_state[axis] is not False:
            raise Exception('Test Failed: Pressing too soon')
        driver.move({axis: points[1]})
        if driver.switch_state[axis] is not True:
            raise Exception('Test Failed: Not hitting switch')
    else:
        try:
            driver.move({axis: points[1]})
            raise Exception('Test Failed: Not hitting switch')
        except SmoothieError:
            pass
    driver.pop_speed()


def _exit_test(driver):
    driver._smoothie_reset()
    driver._setup()
    driver.disengage_axis('XYZABC')


def get_options(parser):
    parser.add_argument(
        "-m",
        "--max_speed",
        dest="max_speed",
        type=int,
        default=100,
        help="Max speed")
    parser.add_argument(
        "-d",
        "--distance",
        dest="distance",
        type=float,
        default=210,
        help="Max distance to travel")
    parser.add_argument(
        "--high_current",
        dest="high_current",
        type=float,
        default=0.5,
        help="Current for the Z stage to run")
    parser.add_argument(
        "-l",
        "--low_current",
        dest="low_current",
        type=float,
        default=0.05,
        help="low z motor current")
    parser.add_argument(
        "-c",
        "--cycles",
        dest="cycles",
        type=int,
        default=2,
        help="Cycles to run test")
    parser.add_argument(
        "-t",
        "--tolerance",
        dest="tolerance",
        type=float,
        default=0.5,
        help="Axis tolerance in millimeters")
    return parser.parse_args()


def run_z_stage(driver, options):
    for cycle in range(options.cycles):
        pick_up_motion(
            options.distance, options.max_speed, 30, driver, options)
        try:
            print("Testing Z")
            test_axis('Z', options.tolerance, driver)
        except Exception as e:
            print("FAIL: {}".format(e))
        try:
            print("Testing A")
            test_axis('A', options.tolerance, driver)
        except Exception as e:
            print("FAIL: {}".format(e))


if __name__ == '__main__':
    parser = args_handler.root_argparser(
        "Test the z-stages of the OT-2")
    args = get_options(parser)
    _, driver = args_handler.build_driver(args.port)
    atexit.register(functools.partial(_exit_test, driver))
    logging.basicConfig(filename='z-stage-test.log')
    try:
        print("In Progress.. ")
        setup(args.high_current, args.max_speed, driver)
        driver.home("ZA")
        run_z_stage(driver, args)
    except KeyboardInterrupt:
        print("Test Cancelled")
        exit()
    except Exception as e:
        print("FAIL: {}".format(e))
        exit()
    print("PASS")



from __future__ import annotations

import asyncio
import logging
from typing import Generic, List, Optional, TYPE_CHECKING, TypeVar, Union

from opentrons import types
from opentrons.hardware_control import modules
from opentrons.hardware_control.types import Axis
from opentrons.commands import module_commands as cmds
from opentrons.commands.publisher import CommandPublisher, publish
from opentrons.protocols.api_support.types import APIVersion

from .labware import (
    Labware, load, load_from_definition)
from opentrons.protocols.geometry.module_geometry import ModuleGeometry,\
    ThermocyclerGeometry
from opentrons.protocols.api_support.util import requires_version

if TYPE_CHECKING:
    from.protocol_context import ProtocolContext
    from opentrons_shared_data.labware.dev_types import (LabwareDefinition)


ModuleTypes = Union[
    'TemperatureModuleContext',
    'MagneticModuleContext',
    'ThermocyclerContext'
]

ENGAGE_HEIGHT_UNIT_CNV = 2
STANDARD_MAGDECK_LABWARE = [
    'biorad_96_wellplate_200ul_pcr',
    'nest_96_wellplate_100ul_pcr_full_skirt',
    'usascientific_96_wellplate_2.4ml_deep']

MODULE_LOG = logging.getLogger(__name__)

GeometryType = TypeVar('GeometryType', bound=ModuleGeometry)
class ModuleContext(CommandPublisher, Generic[GeometryType]):  # noqa: E302
    """ An object representing a connected module.


    .. versionadded:: 2.0

    """

    def __init__(self,
                 ctx: ProtocolContext,
                 geometry: GeometryType,
                 at_version: APIVersion) -> None:
        """ Build the ModuleContext.

        This usually should not be instantiated directly; instead, modules
        should be loaded using :py:meth:`ProtocolContext.load_module`.

        :param ctx: The parent context for the module
        :param geometry: The :py:class:`.ModuleGeometry` for the module
        """
        super().__init__(ctx.broker)
        self._geometry = geometry
        self._ctx = ctx
        self._api_version = at_version

    @property  # type: ignore
    @requires_version(2, 0)
    def api_version(self) -> APIVersion:
        return self._api_version

    @requires_version(2, 0)
    def load_labware_object(self, labware: Labware) -> Labware:
        """ Specify the presence of a piece of labware on the module.

        :param labware: The labware object. This object should be already
                        initialized and its parent should be set to this
                        module's geometry. To initialize and load a labware
                        onto the module in one step, see
                        :py:meth:`load_labware`.
        :returns: The properly-linked labware object
        """
        mod_labware = self._geometry.add_labware(labware)
        self._ctx._implementation.get_deck().recalculate_high_z()
        return mod_labware

    @requires_version(2, 0)
    def load_labware(
            self,
            name: str,
            label: Optional[str] = None,
            namespace: Optional[str] = None,
            version: int = 1,
            ) -> Labware:
        """ Specify the presence of a piece of labware on the module.

        :param name: The name of the labware object.
        :param str label: An optional special name to give the labware. If
            specified, this is the name the labware will appear as in the run
            log and the calibration view in the Opentrons app.
        :param str namespace: The namespace the labware definition belongs to.
            If unspecified, will search 'opentrons' then 'custom_beta'
        :param int version: The version of the labware definition. If
            unspecified, will use version 1.

        :returns: The initialized and loaded labware object.

        .. versionadded:: 2.1
            The *label,* *namespace,* and *version* parameters.
        """
        if self.api_version < APIVersion(2, 1) and\
                (label or namespace or version):
            MODULE_LOG.warning(
                f'You have specified API {self.api_version}, but you '
                'are trying to utilize new load_labware parameters in 2.1')
        lw = load(name, self._geometry.location,
                  label, namespace, version,
                  bundled_defs=self._ctx._implementation.get_bundled_labware(),
                  extra_defs=self._ctx._implementation.get_extra_labware())
        return self.load_labware_object(lw)

    @requires_version(2, 0)
    def load_labware_from_definition(
            self,
            definition: 'LabwareDefinition',
            label: Optional[str] = None) -> Labware:
        """
        Specify the presence of a labware on the module, using an
        inline definition.

        :param definition: The labware definition.
        :param str label: An optional special name to give the labware. If
                          specified, this is the name the labware will appear
                          as in the run log and the calibration view in the
                          Opentrons app.
        :returns: The initialized and loaded labware object.
        """
        lw = load_from_definition(
            definition, self._geometry.location, label)
        return self.load_labware_object(lw)

    @requires_version(2, 1)
    def load_labware_by_name(self,
                             name: str,
                             label: Optional[str] = None,
                             namespace: Optional[str] = None,
                             version: int = 1,) -> Labware:
        """
        .. deprecated:: 2.0
            Use :py:meth:`load_labware` instead.
        """
        MODULE_LOG.warning(
            'load_labware_by_name is deprecated. Use load_labware instead.')
        return self.load_labware(name, label, namespace, version)

    @property  # type: ignore
    @requires_version(2, 0)
    def labware(self) -> Optional[Labware]:
        """ The labware (if any) present on this module. """
        return self._geometry.labware

    @property  # type: ignore
    @requires_version(2, 0)
    def geometry(self) -> ModuleGeometry:
        """ The object representing the module as an item on the deck

        :returns: ModuleGeometry
        """
        return self._geometry

    def __repr__(self):
        return "{} at {} lw {}".format(self.__class__.__name__,
                                       self._geometry,
                                       self.labware)


class TemperatureModuleContext(ModuleContext[ModuleGeometry]):
    """ An object representing a connected Temperature Module.

    It should not be instantiated directly; instead, it should be
    created through :py:meth:`.ProtocolContext.load_module` using:
    ``ctx.load_module('Temperature Module', slot_number)``.

    A minimal protocol with a Temperature module would look like this:

    .. code block:: python

        def run(ctx):
            slot_number = 10
            temp_mod = ctx.load_module('Temperature Module', slot_number)
            temp_plate = temp_mod.load_labware(
                'biorad_96_wellplate_200ul_pcr')

            temp_mod.set_temperature(45.5)
            temp_mod.deactivate()

    .. note::

        In order to prevent physical obstruction of other slots, place the
        Temperature Module in a slot on the horizontal edges of the deck (such
        as 1, 4, 7, or 10 on the left or 3, 6, or 7 on the right), with the USB
        cable and power cord pointing away from the deck.


    .. versionadded:: 2.0

    """
    def __init__(self,
                 ctx: ProtocolContext,
                 hw_module: modules.tempdeck.TempDeck,
                 geometry: ModuleGeometry,
                 at_version: APIVersion,
                 loop: asyncio.AbstractEventLoop) -> None:
        self._module = hw_module
        self._loop = loop
        super().__init__(ctx, geometry, at_version)

    @publish.both(command=cmds.tempdeck_set_temp)
    @requires_version(2, 0)
    def set_temperature(self, celsius: float):
        """ Set the target temperature, in C.

        Must be between 4 and 95C based on Opentrons QA.

        :param celsius: The target temperature, in C
        """
        return self._module.set_temperature(celsius)

    @publish.both(command=cmds.tempdeck_set_temp)
    @requires_version(2, 3)
    def start_set_temperature(self, celsius: float):
        """ Start setting the target temperature, in C.

        Must be between 4 and 95C based on Opentrons QA.

        :param celsius: The target temperature, in C
        """
        return self._module.start_set_temperature(celsius)

    @publish.both(command=cmds.tempdeck_await_temp)
    @requires_version(2, 3)
    def await_temperature(self, celsius: float):
        """ Wait until module reaches temperature, in C.

        Must be between 4 and 95C based on Opentrons QA.

        :param celsius: The target temperature, in C
        """
        return self._module.await_temperature(celsius)

    @publish.both(command=cmds.tempdeck_deactivate)
    @requires_version(2, 0)
    def deactivate(self):
        """ Stop heating (or cooling) and turn off the fan.
        """
        return self._module.deactivate()

    @property  # type: ignore
    @requires_version(2, 0)
    def temperature(self):
        """ Current temperature in C"""
        return self._module.temperature

    @property  # type: ignore
    @requires_version(2, 0)
    def target(self):
        """ Current target temperature in C"""
        return self._module.target

    @property  # type: ignore
    @requires_version(2, 3)
    def status(self):
        """ The status of the module.

        Returns 'holding at target', 'cooling', 'heating', or 'idle'

        """
        return self._module.status


class MagneticModuleContext(ModuleContext[ModuleGeometry]):
    """ An object representing a connected Temperature Module.

    It should not be instantiated directly; instead, it should be
    created through :py:meth:`.ProtocolContext.load_module`.

    .. versionadded:: 2.0

    """
    def __init__(self,
                 ctx: ProtocolContext,
                 hw_module: modules.magdeck.MagDeck,
                 geometry: ModuleGeometry,
                 at_version: APIVersion,
                 loop: asyncio.AbstractEventLoop) -> None:
        self._module = hw_module
        self._loop = loop
        super().__init__(ctx, geometry, at_version)

    @publish.both(command=cmds.magdeck_calibrate)
    @requires_version(2, 0)
    def calibrate(self):
        """ Calibrate the Magnetic Module.

        The calibration is used to establish the position of the lawbare on
        top of the magnetic module.
        """
        self._module.calibrate()

    @requires_version(2, 0)
    def load_labware_object(self, labware: Labware) -> Labware:
        """
        Load labware onto a Magnetic Module, checking if it is compatible
        """
        if labware.magdeck_engage_height is None:
            MODULE_LOG.warning(
                "This labware ({}) is not explicitly compatible with the"
                " Magnetic Module. You will have to specify a height when"
                " calling engage().")
        return super().load_labware_object(labware)

    @publish.both(command=cmds.magdeck_engage)
    @requires_version(2, 0)
    def engage(self,
               height: Optional[float] = None,
               offset: Optional[float] = None,
               height_from_base: Optional[float] = None):
        """ Raise the Magnetic Module's magnets.

        The destination of the magnets can be specified in several different
        ways, based on internally stored default heights for labware:

           - If neither ``height``, ``height_from_base`` nor ``offset`` is
             specified, the magnets will raise to a reasonable default height
             based on the specified labware.
           - The recommended way to adjust the height of the magnets is to
             specify ``height_from_base``, which should be a distance in mm
             relative to the base of the labware that is on the magnetic module
           - If ``height`` is specified, it should be a distance in mm from the
             home position of the magnets.
           - If ``offset`` is specified, it should be an offset in mm from the
             default position. A positive number moves the magnets higher and
             a negative number moves the magnets lower.

        Only certain labwares have defined engage heights for the Magnetic
        Module. If a labware that does not have a defined engage height is
        loaded on the Magnetic Module (or if no labware is loaded), then
        ``height`` or ``height_from_labware`` must be specified.

        :param height_from_base: The height to raise the magnets to, in mm from
                                 the base of the labware
        :param height: The height to raise the magnets to, in mm from home.
        :param offset: An offset relative to the default height for the labware
                       in mm

        .. versionadded:: 2.1
            The *height_from_base* parameter.
        """
        if height is not None:
            dist = height
        elif height_from_base is not None and\
                self._ctx._api_version >= APIVersion(2, 2):
            dist = height_from_base +\
                modules.magdeck.OFFSET_TO_LABWARE_BOTTOM[self._module.model()]
        elif self.labware and self.labware.magdeck_engage_height is not None:
            dist = self._determine_lw_engage_height()
            if offset:
                dist += offset
        else:
            raise ValueError(
                "Currently loaded labware {} does not have a known engage "
                "height; please specify explicitly with the height param"
                .format(self.labware))
        self._module.engage(dist)

    def _determine_lw_engage_height(self) -> float:
        """ Return engage height based on Protocol API and module versions

        For API Version 2.3 or later:
           - Multiply non-standard labware engage heights by 2 for gen1 modules
           - Divide standard labware engage heights by 2 for gen2 modules
        If none of the above, return the labware engage heights as defined in
        the labware definitions
        """
        assert self.labware, self.labware.magdeck_engage_height

        engage_height = self.labware.magdeck_engage_height

        is_api_breakpoint = (self._ctx._api_version >= APIVersion(2, 3))
        is_v1_module = (self._module.model() == 'magneticModuleV1')
        is_standard_lw = self.labware.load_name in STANDARD_MAGDECK_LABWARE

        if is_api_breakpoint and is_v1_module and not is_standard_lw:
            return engage_height * ENGAGE_HEIGHT_UNIT_CNV
        elif is_api_breakpoint and not is_v1_module and is_standard_lw:
            return engage_height / ENGAGE_HEIGHT_UNIT_CNV
        else:
            return engage_height

    @publish.both(command=cmds.magdeck_disengage)
    @requires_version(2, 0)
    def disengage(self):
        """ Lower the magnets back into the Magnetic Module.
        """
        self._module.deactivate()

    @property  # type: ignore
    @requires_version(2, 0)
    def status(self):
        """ The status of the module. either 'engaged' or 'disengaged' """
        return self._module.status


class ThermocyclerContext(ModuleContext[ThermocyclerGeometry]):
    """ An object representing a connected Temperature Module.

    It should not be instantiated directly; instead, it should be
    created through :py:meth:`.ProtocolContext.load_module`.

    .. versionadded:: 2.0
    """
    def __init__(self,
                 ctx: ProtocolContext,
                 hw_module: modules.thermocycler.Thermocycler,
                 geometry: ThermocyclerGeometry,
                 at_version: APIVersion,
                 loop: asyncio.AbstractEventLoop) -> None:
        self._module = hw_module
        self._loop = loop
        super().__init__(ctx, geometry, at_version)

    def _prepare_for_lid_move(self):
        loaded_instruments = [instr for mount, instr in
                              self._ctx._instruments.items()
                              if instr is not None]
        try:
            instr = loaded_instruments[0]
        except IndexError:
            MODULE_LOG.warning(
                "Cannot assure a safe gantry position to avoid colliding"
                " with the lid of the Thermocycler Module.")
        else:
            ctx_impl = self._ctx._implementation
            instr_impl = instr._implementation
            hardware = ctx_impl.get_hardware().hardware

            hardware.retract(instr_impl.get_mount())
            high_point = hardware.current_position(instr_impl.get_mount())
            trash_top = self._ctx.fixed_trash.wells()[0].top()
            safe_point = trash_top.point._replace(
                    z=high_point[Axis.by_mount(instr_impl.get_mount())])
            instr.move_to(types.Location(safe_point, None), force_direct=True)

    def flag_unsafe_move(self,
                         to_loc: types.Location,
                         from_loc: types.Location):
        to_lw, to_well = to_loc.labware.get_parent_labware_and_well()
        from_lw, from_well = from_loc.labware.get_parent_labware_and_well()
        if self.labware is not None and \
                (self.labware == to_lw or self.labware == from_lw) and \
                self.lid_position != 'open':
            raise RuntimeError(
                "Cannot move to labware loaded in Thermocycler"
                " when lid is not fully open.")

    @publish.both(command=cmds.thermocycler_open)
    @requires_version(2, 0)
    def open_lid(self):
        """ Opens the lid"""
        self._prepare_for_lid_move()
        self._geometry.lid_status = self._module.open()  # type: ignore
        return self._geometry.lid_status

    @publish.both(command=cmds.thermocycler_close)
    @requires_version(2, 0)
    def close_lid(self):
        """ Closes the lid"""
        self._prepare_for_lid_move()
        self._geometry.lid_status = self._module.close()  # type: ignore
        return self._geometry.lid_status

    @publish.both(command=cmds.thermocycler_set_block_temp)
    @requires_version(2, 0)
    def set_block_temperature(self,
                              temperature: float,
                              hold_time_seconds: Optional[float] = None,
                              hold_time_minutes: Optional[float] = None,
                              ramp_rate: Optional[float] = None,
                              block_max_volume: Optional[float] = None):
        """ Set the target temperature for the well block, in °C.

        Valid operational range yet to be determined.

        :param temperature: The target temperature, in °C.
        :param hold_time_minutes: The number of minutes to hold, after reaching
                                  ``temperature``, before proceeding to the
                                  next command.
        :param hold_time_seconds: The number of seconds to hold, after reaching
                                  ``temperature``, before proceeding to the
                                  next command. If ``hold_time_minutes`` and
                                  ``hold_time_seconds`` are not specified,
                                  the Thermocycler will proceed to the next
                                  command after ``temperature`` is reached.
        :param ramp_rate: The target rate of temperature change, in °C/sec.
                          If ``ramp_rate`` is not specified, it will default
                          to the maximum ramp rate as defined in the device
                          configuration.
        :param block_max_volume: The maximum volume of any individual well
                                 of the loaded labware. If not supplied,
                                 the thermocycler will default to 25µL/well.

        .. note:

            If ``hold_time_minutes`` and ``hold_time_seconds`` are not
            specified, the Thermocycler will proceed to the next command
            after ``temperature`` is reached.
        """

        return self._module.set_temperature(
                temperature=temperature,
                hold_time_seconds=hold_time_seconds,
                hold_time_minutes=hold_time_minutes,
                ramp_rate=ramp_rate,
                volume=block_max_volume)

    @publish.both(command=cmds.thermocycler_set_lid_temperature)
    @requires_version(2, 0)
    def set_lid_temperature(self, temperature: float):
        """ Set the target temperature for the heated lid, in °C.

        :param temperature: The target temperature, in °C clamped to the
                            range 20°C to 105°C.

        .. note:

            The Thermocycler will proceed to the next command after
            ``temperature`` has been reached.

        """
        self._module.set_lid_temperature(temperature)

    @publish.both(command=cmds.thermocycler_execute_profile)
    @requires_version(2, 0)
    def execute_profile(self,
                        steps: List[modules.ThermocyclerStep],
                        repetitions: int,
                        block_max_volume: Optional[float] = None):
        """ Execute a Thermocycler Profile defined as a cycle of
        :py:attr:`steps` to repeat for a given number of :py:attr:`repetitions`

        :param steps: List of unique steps that make up a single cycle.
                      Each list item should be a dictionary that maps to
                      the parameters of the :py:meth:`set_block_temperature`
                      method with keys 'temperature', 'hold_time_seconds',
                      and 'hold_time_minutes'.
        :param repetitions: The number of times to repeat the cycled steps.
        :param block_max_volume: The maximum volume of any individual well
                                 of the loaded labware. If not supplied,
                                 the thermocycler will default to 25µL/well.

        .. note:

            Unlike the :py:meth:`set_block_temperature`, either or both of
            'hold_time_minutes' and 'hold_time_seconds' must be defined
            and finite for each step.

        """
        if repetitions <= 0:
            raise ValueError("repetitions must be a positive integer")
        for step in steps:
            if step.get('temperature') is None:
                raise ValueError(
                        "temperature must be defined for each step in cycle")
            hold_mins = step.get('hold_time_minutes')
            hold_secs = step.get('hold_time_seconds')
            if hold_mins is None and hold_secs is None:
                raise ValueError(
                        "either hold_time_minutes or hold_time_seconds must be"
                        "defined for each step in cycle")
        return self._module.cycle_temperatures(steps=steps,
                                               repetitions=repetitions,
                                               volume=block_max_volume)

    @publish.both(command=cmds.thermocycler_deactivate_lid)
    @requires_version(2, 0)
    def deactivate_lid(self):
        """ Turn off the heated lid """
        self._module.deactivate_lid()

    @publish.both(command=cmds.thermocycler_deactivate_block)
    @requires_version(2, 0)
    def deactivate_block(self):
        """ Turn off the well block temperature controller"""
        self._module.deactivate_block()

    @publish.both(command=cmds.thermocycler_deactivate)
    @requires_version(2, 0)
    def deactivate(self):
        """ Turn off the well block temperature controller, and heated lid """
        self._module.deactivate()

    @property  # type: ignore
    @requires_version(2, 0)
    def lid_position(self):
        """ Lid open/close status string"""
        return self._module.lid_status

    @property  # type: ignore
    @requires_version(2, 0)
    def block_temperature_status(self):
        return self._module.status

    @property  # type: ignore
    @requires_version(2, 0)
    def lid_temperature_status(self):
        return self._module.lid_temp_status

    @property  # type: ignore
    @requires_version(2, 0)
    def block_temperature(self):
        """ Current temperature in degrees C """
        return self._module.temperature

    @property  # type: ignore
    @requires_version(2, 0)
    def block_target_temperature(self):
        """ Target temperature in degrees C """
        return self._module.target

    @property  # type: ignore
    @requires_version(2, 0)
    def lid_temperature(self):
        """ Current temperature in degrees C """
        return self._module.lid_temp

    @property  # type: ignore
    @requires_version(2, 0)
    def lid_target_temperature(self):
        """ Target temperature in degrees C """
        return self._module.lid_target

    @property  # type: ignore
    @requires_version(2, 0)
    def ramp_rate(self):
        """ Current ramp rate in degrees C/sec"""
        return self._module.ramp_rate

    @property  # type: ignore
    @requires_version(2, 0)
    def hold_time(self):
        """ Remaining hold time in sec"""
        return self._module.hold_time

    @property  # type: ignore
    @requires_version(2, 0)
    def total_cycle_count(self):
        """ Number of repetitions for current set cycle"""
        return self._module.total_cycle_count

    @property  # type: ignore
    @requires_version(2, 0)
    def current_cycle_index(self):
        """ Index of the current set cycle repetition"""
        return self._module.current_cycle_index

    @property  # type: ignore
    @requires_version(2, 0)
    def total_step_count(self):
        """ Number of steps within the current cycle"""
        return self._module.total_step_count

    @property  # type: ignore
    @requires_version(2, 0)
    def current_step_index(self):
        """ Index of the current step within the current cycle"""
        return self._module.current_step_index



"""
functions and utilities for handling zipped protocol bundles
"""
from datetime import date
import json
from pathlib import PurePosixPath, PurePath
from typing import Dict, BinaryIO, TYPE_CHECKING
from zipfile import ZipFile

from opentrons.calibration_storage.helpers import uri_from_definition

from .types import BundleContents

if TYPE_CHECKING:
    from opentrons_shared_data.labware.dev_types import LabwareDefinition

MAIN_PROTOCOL_FILENAME = 'protocol.ot2.py'
LABWARE_DIR = 'labware'
DATA_DIR = 'data'


def _has_files_at_root(zipFile):
    for zipInfo in zipFile.infolist():
        if zipInfo.filename.count('/') == 0:
            return True
    return False


def extract_bundle(bundle: ZipFile) -> BundleContents:  # noqa: C901
    """ Extract a bundle and verify its contents and structure. """
    if not _has_files_at_root(bundle):
        raise RuntimeError(
            'No files found in ZIP file\'s root directory. When selecting '
            'files to zip, make sure to directly select the files '
            'themselves. Do not select their parent directory, which would '
            'result in nesting all files inside that directory in the ZIP.')
    try:
        with bundle.open(MAIN_PROTOCOL_FILENAME, 'r') as protocol_file:
            py_protocol = protocol_file.read().decode('utf-8')
    except KeyError:
        raise RuntimeError(
            f'Bundled protocol should have a {MAIN_PROTOCOL_FILENAME} ' +
            'file in the root directory')
    bundled_labware: Dict[str, 'LabwareDefinition'] = {}
    bundled_data = {}
    bundled_python = {}
    for zipInfo in bundle.infolist():
        filepath = PurePosixPath(zipInfo.filename)
        rootpath = filepath.parts[0]

        # skip directories and weird OS-added directories
        # (note: the __MACOSX dir would contain '__MACOSX/foo.py'
        # and other files. This would break our inferences, so we need
        # to exclude all contents of that directory)
        if rootpath == '__MACOSX' or zipInfo.is_dir():
            continue

        with bundle.open(zipInfo) as f:
            if rootpath == LABWARE_DIR and filepath.suffix == '.json':
                labware_def = json.loads(f.read().decode('utf-8'))
                labware_key = uri_from_definition(labware_def)
                if labware_key in bundled_labware:
                    raise RuntimeError(
                        f'Conflicting labware in bundle: {labware_key}')
                bundled_labware[labware_key] = labware_def
            elif rootpath == DATA_DIR:
                # note: data files are read as binary
                bundled_data[str(filepath.relative_to(DATA_DIR))] = f.read()
            elif (filepath.suffix == '.py' and
                  str(filepath) != MAIN_PROTOCOL_FILENAME):
                bundled_python[str(filepath)] = f.read().decode('utf-8')

    if not bundled_labware:
        raise RuntimeError('No labware definitions found in bundle.')

    return BundleContents(py_protocol, bundled_labware,
                          bundled_data, bundled_python)


def create_bundle(contents: BundleContents,
                  into_file: BinaryIO):
    """ Create a bundle from assumed-good contents """
    with ZipFile(into_file, mode='w') as zf:
        zf.writestr(MAIN_PROTOCOL_FILENAME, contents.protocol)
        for dataname, datafile in contents.bundled_data.items():
            name = PurePath(dataname).name
            zf.writestr(f'{DATA_DIR}/{name}', datafile)
        for lwdef in contents.bundled_labware.values():
            zipsafe = uri_from_definition(lwdef, '-')
            zf.writestr(f'{LABWARE_DIR}/{zipsafe}.json', json.dumps(lwdef))
        zf.writestr('.bundle_beta', str(date.today()))



""" Constants for the :py:class:`.hardware_control.API` instances.
"""
SHAKE_OFF_TIPS_SPEED = 50  # mm/s
SHAKE_OFF_TIPS_DROP_DISTANCE = 2.25  # mm
SHAKE_OFF_TIPS_PICKUP_DISTANCE = 0.3  # mm
DROP_TIP_RELEASE_DISTANCE = 20  # mm



"""
This was initially generated by datamodel-codegen from the labware schema in
shared-data. It's been modified by hand to be more friendly.

TODO: 20210330 Amit - consider moving this to opentrons-shared-data.
"""

from __future__ import annotations

from enum import Enum
from typing import Dict, List, Optional

from pydantic import BaseModel, Extra, Field
from typing_extensions import Literal

SAFE_STRING_REGEX = '^[a-z0-9._]+$'


class CornerOffsetFromSlot(BaseModel):
    """
    Distance from left-front-bottom corner of slot to left-front-bottom corner
     of labware bounding box. Used for labware that spans multiple slots. For
      labware that does not span multiple slots, x/y/z should all be zero.
    """

    x: float
    y: float
    z: float


class BrandData(BaseModel):
    brand: str = Field(..., description='Brand/manufacturer name')
    brandId: Optional[List[str]] = Field(
        None,
        description='An array of manufacture numbers pertaining to a given labware',
    )
    links: Optional[List[str]] = Field(
        None, description='URLs for manufacturer page(s)'
    )


class DisplayCategory(str, Enum):
    tipRack = 'tipRack'
    tubeRack = 'tubeRack'
    reservoir = 'reservoir'
    trash = 'trash'
    wellPlate = 'wellPlate'
    aluminumBlock = 'aluminumBlock'
    other = 'other'


class Metadata(BaseModel):
    """
    Properties used for search and display
    """

    displayName: str = Field(
        ...,
        description='Easy to remember name of labware')
    displayCategory: DisplayCategory = Field(
        ...,
        description='Label(s) used in UI to categorize labware'
    )
    displayVolumeUnits: Literal['µL', 'mL', 'L'] = Field(
        ...,
        description='Volume units for display'
    )
    tags: Optional[List[str]] = Field(
        None,
        description='List of descriptions for a given labware'
    )


class Parameters(BaseModel):
    """
    Internal describers used to determine pipette movement to labware
    """

    format: Literal[
        '96Standard', '384Standard', 'trough', 'irregular', 'trash'
    ] = Field(
        ...,
        description='Property to determine compatibility with multichannel pipette'
    )
    quirks: Optional[List[str]] = Field(
        None,
        description='Property to classify a specific behavior this labware '
                    'should have',
    )
    isTiprack: bool = Field(
        ...,
        description='Flag marking whether a labware is a tiprack or not'
    )
    tipLength: Optional[float] = Field(
        None,
        ge=0.0,
        description='Required if labware is tiprack, specifies length of tip'
                    ' from drawing or as measured with calipers',
    )
    tipOverlap: Optional[float] = Field(
        None,
        ge=0.0,
        description='Required if labware is tiprack, specifies the length of '
                    'the area of the tip that overlaps the nozzle of the pipette',
    )
    loadName: str = Field(
        ...,
        description='Name used to reference a labware definition',
        regex=SAFE_STRING_REGEX
    )
    isMagneticModuleCompatible: bool = Field(
        ...,
        description='Flag marking whether a labware is compatible by default '
                    'with the Magnetic Module',
    )
    magneticModuleEngageHeight: Optional[float] = Field(
        None,
        ge=0.0,
        description='Distance to move magnetic module magnets to engage'
    )


class Dimensions(BaseModel):
    """
    Outer dimensions of a labware
    """

    yDimension: float = Field(..., ge=0.0)
    zDimension: float = Field(..., ge=0.0)
    xDimension: float = Field(..., ge=0.0)


class WellDefinition(BaseModel):
    class Config:
        extra = Extra.allow

    depth: float = Field(..., ge=0.0)
    x: float = Field(
        ...,
        ge=0.0,
        description='x location of center-bottom of well in reference to '
                    'left-front-bottom of labware',
    )
    y: float = Field(
        ...,
        ge=0.0,
        description='y location of center-bottom of well in reference to '
                    'left-front-bottom of labware',
    )
    z: float = Field(
        ...,
        ge=0.0,
        description='z location of center-bottom of well in reference to '
                    'left-front-bottom of labware',
    )
    totalLiquidVolume: float = Field(
        ...,
        ge=0.0,
        description='Total well, tube, or tip volume in microliters'
    )
    xDimension: Optional[float] = Field(
        None,
        ge=0.0,
        description='x dimension of rectangular wells'
    )
    yDimension: Optional[float] = Field(
        None,
        ge=0.0,
        description='y dimension of rectangular wells'
    )
    diameter: Optional[float] = Field(
        None,
        ge=0.0,
        description='diameter of circular wells'
    )
    shape: Literal['rectangular', 'circular'] = Field(
        ...,
        description="If 'rectangular', use xDimension and "
                    "yDimension; if 'circular' use diameter",
    )


class Metadata1(BaseModel):
    """
    Metadata specific to a grid of wells in a labware
    """

    displayName: Optional[str] = Field(
        None, description='User-readable name for the well group'
    )
    displayCategory: Optional[DisplayCategory] = Field(
        None, description='Label(s) used in UI to categorize well groups'
    )
    wellBottomShape: Optional[Literal['flat', 'u', 'v']] = Field(
        None, description='Bottom shape of the well for UI purposes'
    )


class Group(BaseModel):
    wells: List[str] = Field(
        ..., description='An array of wells that contain the same metadata',
        min_items=1
    )
    metadata: Metadata1 = Field(
        ..., description='Metadata specific to a grid of wells in a labware'
    )
    brand: Optional[BrandData] = Field(
        None, description='Brand data for the well group (e.g. for tubes)'
    )


class LabwareDefinition(BaseModel):
    schemaVersion: Literal[1, 2] = Field(
        ..., description='Which schema version a labware is using'
    )
    version: int = Field(
        ...,
        description='Version of the labware definition itself '
                    '(eg myPlate v1/v2/v3). An incrementing integer',
        ge=1.0,
    )
    namespace: str = Field(..., regex=SAFE_STRING_REGEX)
    metadata: Metadata = Field(
        ..., description='Properties used for search and display'
    )
    brand: BrandData = Field(
        ...,
        description='Real-world labware that the definition is modeled '
                    'from and/or compatible with',
    )
    parameters: Parameters = Field(
        ...,
        description='Internal describers used to determine pipette movement '
                    'to labware',
    )
    ordering: List[List[str]] = Field(
        ...,
        description='Generated array that keeps track of how wells should be '
                    'ordered in a labware',
    )
    cornerOffsetFromSlot: CornerOffsetFromSlot = Field(
        ...,
        description='Distance from left-front-bottom corner of slot to '
                    'left-front-bottom corner of labware bounding box. Used for '
                    'labware that spans multiple slots. For labware that does '
                    'not span multiple slots, x/y/z should all be zero.',
    )
    dimensions: Dimensions = Field(
        ...,
        description='Outer dimensions of a labware'
    )
    wells: Dict[str, WellDefinition] = Field(
        ...,
        description='Unordered object of well objects with position and '
                    'dimensional information',
    )
    groups: List[Group] = Field(
        ...,
        description='Logical well groupings for metadata/display purposes; '
                    'changes in groups do not affect protocol execution',
    )



"""Dispense command request, result, and implementation models."""
from __future__ import annotations
from typing_extensions import final

from .command import CommandImplementation, CommandHandlers
from .pipetting_common import BaseLiquidHandlingRequest, BaseLiquidHandlingResult


@final
class DispenseRequest(BaseLiquidHandlingRequest):
    """A request to aspirate from a specific well."""

    def get_implementation(self) -> DispenseImplementation:
        """Get the execution implementation of the DispenseRequest."""
        return DispenseImplementation(self)


@final
class DispenseResult(BaseLiquidHandlingResult):
    """Result data from the execution of a DispenseRequest."""

    pass


@final
class DispenseImplementation(
    CommandImplementation[DispenseRequest, DispenseResult]
):
    """Dispense command implementation."""

    async def execute(self, handlers: CommandHandlers) -> DispenseResult:
        """Move to and dispense to the requested well."""
        volume = await handlers.pipetting.dispense(
            pipette_id=self._request.pipetteId,
            labware_id=self._request.labwareId,
            well_name=self._request.wellName,
            well_location=self._request.wellLocation,
            volume=self._request.volume,
        )

        return DispenseResult(volume=volume)



"""Load labware command request, result, and implementation models."""
from __future__ import annotations

from pydantic import BaseModel, Field
from typing import Tuple, Optional

from opentrons.protocols.models import LabwareDefinition

from ..types import LabwareLocation
from .command import CommandImplementation, CommandHandlers


class LoadLabwareRequest(BaseModel):
    """A request to load a labware into a slot."""

    location: LabwareLocation = Field(
        ...,
        description="Location the labware should be loaded into.",
    )
    loadName: str = Field(
        ...,
        description="Name used to reference a labware definition.",
    )
    namespace: str = Field(
        ...,
        description="The namespace the labware definition belongs to.",
    )
    version: int = Field(
        ...,
        description="The labware definition version.",
    )
    labwareId: Optional[str] = Field(
        None,
        description="An optional ID to assign to this labware. If None, an ID "
                    "will be generated."
    )

    def get_implementation(self) -> LoadLabwareImplementation:
        """Get the load labware request's command implementation."""
        return LoadLabwareImplementation(self)


class LoadLabwareResult(BaseModel):
    """Result data from the execution of a LoadLabwareRequest."""

    labwareId: str = Field(
        ...,
        description="An ID to reference this labware in subsequent commands.",
    )
    definition: LabwareDefinition = Field(
        ...,
        description="The full definition data for this labware.",
    )
    calibration: Tuple[float, float, float] = Field(
        ...,
        description="Calibration offset data for this labware at load time.",
    )


class LoadLabwareImplementation(
    CommandImplementation[LoadLabwareRequest, LoadLabwareResult]
):
    """Load labware command implementation."""

    async def execute(self, handlers: CommandHandlers) -> LoadLabwareResult:
        """Load definition and calibration data necessary for a labware."""
        loaded_labware = await handlers.equipment.load_labware(
            load_name=self._request.loadName,
            namespace=self._request.namespace,
            version=self._request.version,
            location=self._request.location,
            labware_id=self._request.labwareId
        )

        return LoadLabwareResult(
            labwareId=loaded_labware.labware_id,
            definition=loaded_labware.definition,
            calibration=loaded_labware.calibration,
        )



from __future__ import annotations

import logging
from functools import partial
from typing import TYPE_CHECKING, Union, Optional, Callable, Tuple

from opentrons import types
from opentrons.hardware_control.types import CriticalPoint
from opentrons.protocol_api.module_contexts import ThermocyclerContext
from opentrons.protocol_api.labware import (
    Labware, Well)
from opentrons.protocols.api_support.labware_like import LabwareLike
from opentrons.protocols.geometry import planning
from opentrons.protocols.api_support.util import build_edges

if TYPE_CHECKING:
    from opentrons.protocol_api.protocol_context import ProtocolContext
    from opentrons.protocol_api.instrument_context import InstrumentContext
    from opentrons.hardware_control import types as hc_types
    from opentrons.protocols.api_support.util import HardwareManager


class PairedInstrument:

    def __init__(self,
                 primary_instrument: InstrumentContext,
                 secondary_instrument: InstrumentContext,
                 pair_policy: hc_types.PipettePair,
                 ctx: ProtocolContext,
                 hardware_manager: HardwareManager,
                 log_parent: logging.Logger):
        self.p_instrument = primary_instrument
        self.s_instrument = secondary_instrument
        self._pair_policy = pair_policy
        self._ctx = ctx
        self._hw_manager = hardware_manager
        self._log = log_parent.getChild(repr(self))

        self._last_location: Union[Labware, Well, None] = None

    def pick_up_tip(self, target: Well, secondary_target: Well,
                    tiprack: Labware, presses: Optional[int],
                    increment: Optional[float], tip_length: float):

        self.move_to(target.top())

        self._hw_manager.hardware.set_current_tiprack_diameter(
            self._pair_policy, target.diameter)
        self._hw_manager.hardware.pick_up_tip(
            self._pair_policy, tip_length, presses, increment)

        self._hw_manager.hardware.set_working_volume(
            self._pair_policy, target.max_volume)

        tiprack.use_tips(target, self.p_instrument.channels)
        tiprack.use_tips(secondary_target, self.s_instrument.channels)

    def drop_tip(self, target: types.Location, home_after: bool):
        self.move_to(target)
        self._hw_manager.hardware.drop_tip(
            self._pair_policy, home_after=home_after)
        return self

    def move_to(self, location: types.Location, force_direct: bool = False,
                minimum_z_height: Optional[float] = None,
                speed: Optional[float] = None):
        if not speed:
            speed = self.p_instrument.default_speed

        last_location = self._ctx.location_cache
        if last_location:
            from_lw = last_location.labware
        else:
            from_lw = LabwareLike(None)

        from_center = 'centerMultichannelOnWells'\
            in from_lw.quirks_from_any_parent()
        cp_override = CriticalPoint.XY_CENTER if from_center else None
        from_loc = types.Location(
            self._hw_manager.hardware.gantry_position(
                self._pair_policy.primary, critical_point=cp_override),
            from_lw)

        for mod in self._ctx._modules:
            if isinstance(mod, ThermocyclerContext):
                mod.flag_unsafe_move(to_loc=location, from_loc=from_loc)

        primary_height = \
            self._hw_manager.hardware.get_instrument_max_height(
                self._pair_policy.primary)
        secondary_height = \
            self._hw_manager.hardware.get_instrument_max_height(
                self._pair_policy.secondary)
        moves = planning.plan_moves(from_loc,
                                    location,
                                    self._ctx._implementation.get_deck(),
                                    min(primary_height, secondary_height),
                                    force_direct=force_direct,
                                    minimum_z_height=minimum_z_height
                                    )
        self._log.debug("move_to: {}->{} via:\n\t{}"
                        .format(from_loc, location, moves))
        try:
            for move in moves:
                self._hw_manager.hardware.move_to(
                    self._pair_policy, move[0], critical_point=move[1],
                    speed=speed,
                    max_speeds=self._ctx._implementation.get_max_speeds().data)
        except Exception:
            self._ctx.location_cache = None
            raise
        else:
            self._ctx.location_cache = location
        return self

    def aspirate(
            self, volume: Optional[float],
            location: Optional[types.Location] = None,
            rate: Optional[float] = 1.0) -> Tuple[types.Location, Callable]:
        last_location = self._ctx.location_cache
        if location:
            loc = location
        elif last_location:
            loc = last_location
        else:
            raise RuntimeError(
                "If aspirate is called without an explicit location, another"
                " method that moves to a location (such as move_to or "
                "dispense) must previously have been called so the robot "
                "knows where it is.")

        if self.p_instrument.current_volume == 0:
            # Make sure we're at the top of the labware and clear of any
            # liquid to prepare the pipette for aspiration
            primary_ready = self.p_instrument.hw_pipette['ready_to_aspirate']
            secondary_ready = self.s_instrument.hw_pipette['ready_to_aspirate']
            if not primary_ready or not secondary_ready:
                if loc.labware.is_well:
                    self.move_to(loc.labware.as_well().top())
                else:
                    # TODO(seth,2019/7/29): This should be a warning exposed
                    #  via rpc to the runapp
                    self._log.warning(
                        "When aspirate is called on something other than a "
                        "well relative position, we can't move to the top of"
                        " the well to prepare for aspiration. This might "
                        "cause over aspiration if the previous command is a "
                        "blow_out.")
                self._hw_manager.hardware.prepare_for_aspirate(
                    self._pair_policy)
            self.move_to(loc)
        elif loc != last_location:
            self.move_to(loc)

        hw_aspirate = self._hw_manager.hardware.aspirate
        return loc, partial(hw_aspirate, self._pair_policy, volume, rate)

    def dispense(
            self, volume: Optional[float],
            location: Optional[types.Location],
            rate: float) -> Tuple[types.Location, Callable]:
        last_location = self._ctx.location_cache
        if location:
            loc = location
            self.move_to(location)
        elif last_location:
            loc = last_location
            pass
        else:
            raise RuntimeError(
                "If dispense is called without an explicit location, another"
                " method that moves to a location (such as move_to or "
                "aspirate) must previously have been called so the robot "
                "knows where it is.")
        hw_dispense = self._hw_manager.hardware.dispense
        return loc, partial(hw_dispense, self._pair_policy, volume, rate)

    def blow_out(self, location: types.Location):
        if location:
            self.move_to(location)
        elif self._ctx.location_cache:
            # if location cache exists, pipette blows out immediately at
            # current location, no movement is needed
            pass
        else:
            raise RuntimeError(
                "If blow out is called without an explicit location, another"
                " method that moves to a location (such as move_to or "
                "dispense) must previously have been called so the robot "
                "knows where it is.")
        self._hw_manager.hardware.blow_out(self._pair_policy)

    def air_gap(self, volume: Optional[float], height: float):
        loc = self._ctx.location_cache
        if not loc or not loc.labware.is_well:
            raise RuntimeError('No previous Well cached to perform air gap')
        target = loc.labware.as_well().top(height)
        self.move_to(target)
        # Aspirate now returns a partial function for run log purposes
        # but air gap only cares about executing an aspirate so here
        # we automatically call the partial function
        self.aspirate(volume)[1]()

    def touch_tip(
            self, location: Optional[Well], radius: float,
            v_offset: float, speed: float):
        if location is None:
            if not self._ctx.location_cache:
                raise RuntimeError('No valid current location cache present')
            else:
                well = self._ctx.location_cache.labware  # type: ignore
                # type checked below
        else:
            well = LabwareLike(location)

        if well.is_well:
            if 'touchTipDisabled' in well.quirks_from_any_parent():
                self._log.info(f"Ignoring touch tip on labware {well}")
                return self
            if well.parent.as_labware().is_tiprack:
                self._log.warning('Touch_tip being performed on a tiprack. '
                                  'Please re-check your code')

            move_with_z_offset =\
                well.as_well().top().point + types.Point(0, 0, v_offset)
            to_loc = types.Location(move_with_z_offset, well)
            self.move_to(to_loc)
        else:
            # If location is a not a valid well, raise a type error
            raise TypeError(
                'location should be a Well, but it is {}'.format(location))

        edges = build_edges(
            well.as_well(), v_offset, self._pair_policy.primary,
            self._ctx._implementation.get_deck(), radius)
        for edge in edges:
            self._hw_manager.hardware.move_to(
                self._pair_policy, edge, speed)



import asyncio
import atexit
import functools
import logging
import os
import subprocess
from typing import Dict, TYPE_CHECKING

from opentrons.config import infer_config_base_dir
from opentrons.drivers import serial_communication
if TYPE_CHECKING:
    from opentrons.drivers.smoothie_drivers.driver_3_0\
        import SmoothieDriver_3_0_0
from opentrons.system import nmcli
from . import args_handler

log = logging.getLogger(__name__)

RESULT_SPACE = '\t- {}'
FAIL = 'FAIL\t*** !!! ***'
PASS = 'PASS'

USB_MOUNT_FILEPATH = '/mnt/usbdrive'
DATA_FOLDER = str(infer_config_base_dir())
VIDEO_FILEPATH = os.path.join(DATA_FOLDER, 'cam_test.jpg')
AUDIO_FILE_PATH = '/etc/audio/speaker-test.mp3'


def _find_storage_device():
    if os.path.exists(USB_MOUNT_FILEPATH) is False:
        run_quiet_process('mkdir {}'.format(USB_MOUNT_FILEPATH))
    if os.path.ismount(USB_MOUNT_FILEPATH) is False:
        sdn1_devices = [
            '/dev/sd{}1'.format(l)
            for l in 'abcdefgh'  # noqa: E741
            if os.path.exists('/dev/sd{}1'.format(l))
        ]
        if len(sdn1_devices) == 0:
            print(RESULT_SPACE.format(FAIL))
            return
        try:
            run_quiet_process(
                'mount {0} {1}'.format(sdn1_devices[0], USB_MOUNT_FILEPATH))
        except Exception:
            print(RESULT_SPACE.format(FAIL))
            return
    return True


def _this_wifi_ip_address():
    loop = asyncio.get_event_loop()
    wifi_info = loop.run_until_complete(
        nmcli.iface_info(nmcli.NETWORK_IFACES.WIFI))
    assert wifi_info['ipAddress'], 'Not connected to wifi'
    return wifi_info['ipAddress'].split('/')[0]


def _erase_data(filepath):
    if os.path.exists(filepath):
        os.remove(filepath)


def _reset_lights(driver: 'SmoothieDriver_3_0_0'):
    driver.turn_off_rail_lights()
    driver.turn_on_blue_button_light()


def _get_state_of_inputs(driver: 'SmoothieDriver_3_0_0'):
    smoothie_switches = driver.switch_state
    probe = smoothie_switches['Probe']
    endstops = {
        ax: val
        for ax, val in smoothie_switches.items()
        if ax in 'XYZA'  # only test gantry axes
    }
    return {
        'button': driver.read_button(),
        'windows': driver.read_window_switches(),
        'probe': probe,
        'endstops': endstops
    }


def _set_lights(state: Dict[str, Dict[str, bool]],
                driver: 'SmoothieDriver_3_0_0'):
    if state['windows']:
        driver.turn_off_rail_lights()
    else:
        driver.turn_on_rail_lights()
    red, green, blue = (False, False, False)
    if any(state['endstops'].values()):
        red = True
    if state['probe']:
        green = True
    if state['button']:
        blue = True
    driver._gpio_chardev.set_button_light(red=red, green=green, blue=blue)


def run_quiet_process(command):
    subprocess.check_output('{} &> /dev/null'.format(command), shell=True)


def _get_unique_smoothie_responses(responses):
    """ Find the number of truly unique responses from the smoothie, ignoring

    - Responses that are only different because of interjected \r\n
    - Responses that are only different by \r or \n replacing a single char

    Both of these errors are results of race conditions between smoothie
    terminal echo mode and responses, and do not indicate serial failures.
    """
    uniques = list(set(responses))  # Eliminate exact repetitions
    # eliminate "uniques" that are really just the second \r\n racing
    # with the smoothie's return values
    true_uniques = [uniques[0]]
    for unique in uniques[1:]:
        if len(unique) != len(uniques[0]):
            true_uniques.append(unique)
            continue
        for a, b in zip(uniques[0], unique):
            if a in '\r\n':
                continue
            elif b in '\r\n':
                continue
            elif a != b:
                true_uniques.append(unique)
                break
    return true_uniques


def test_smoothie_gpio(driver: 'SmoothieDriver_3_0_0'):
    assert driver._connection, 'must be connected'

    def _write_and_return(msg):
        return serial_communication.write_and_return(
            msg + '\r\n',
            'ok\r\n',
            driver._connection,
            timeout=1)

    print('CONNECT')
    # make sure the driver is currently working as expected
    version_response = _write_and_return('version')
    if 'version' in version_response:
        print(RESULT_SPACE.format(PASS))
    else:
        print(RESULT_SPACE.format(FAIL))

    print('DATA LOSS')
    [_write_and_return('version') for i in range(10)]
    # check that if we write the same thing to the smoothie a bunch
    # it works
    data = [_write_and_return('version') for i in range(100)]
    true_uniques = _get_unique_smoothie_responses(data)
    if len(true_uniques) == 1:
        print(RESULT_SPACE.format(PASS))
    else:
        log.info(f'true uniques: {true_uniques}')
        print(RESULT_SPACE.format(FAIL))

    print('HALT')
    driver._connection.reset_input_buffer()
    # drop the HALT line LOW, and make sure there is an error state
    driver.hard_halt()

    old_timeout = int(driver._connection.timeout)
    driver._connection.timeout = 1
    r = driver._connection.readline().decode()
    if 'ALARM' in r:
        print(RESULT_SPACE.format(PASS))
    else:
        print(RESULT_SPACE.format(FAIL))

    driver._reset_from_error()
    driver._connection.timeout = old_timeout

    print('ISP')
    # drop the ISP line to LOW, and make sure it is dead
    driver._smoothie_programming_mode()
    try:
        _write_and_return('M999')
        print(RESULT_SPACE.format(FAIL))
    except Exception:
        print(RESULT_SPACE.format(PASS))

    print('RESET')
    # toggle the RESET line to LOW, and make sure it is NOT dead
    driver._smoothie_reset()
    r = _write_and_return('M119')
    if 'X_max' in r:
        print(RESULT_SPACE.format(PASS))
    else:
        print(RESULT_SPACE.format(FAIL))


def test_switches_and_lights(driver: 'SmoothieDriver_3_0_0'):
    print('\n')
    print('* BUTTON\t--> BLUE')
    print('* PROBE\t\t--> GREEN')
    print('* ENDSTOP\t--> RED')
    print('* WINDOW\t--> LIGHTS')
    print('')
    print('Next\t--> CTRL-C')
    print('')
    # enter button-read loop

    try:
        while True:
            state = _get_state_of_inputs(driver)
            _set_lights(state, driver)
    except KeyboardInterrupt:
        print()
        pass


def test_speaker():
    print('Speaker')
    print('Next\t--> CTRL-C')
    try:
        run_quiet_process('mpg123 {}'.format(AUDIO_FILE_PATH))
    except KeyboardInterrupt:
        pass
        print()


def record_camera(filepath):
    print('USB Camera')
    # record 1 second of video from the USB camera
    c = 'ffmpeg -f video4linux2 -s 640x480 -i /dev/video0 -ss 0:0:1 -frames 1 {}'  # noqa: E501
    try:
        run_quiet_process(c.format(filepath))
        print(RESULT_SPACE.format(PASS))
    except Exception:
        print(RESULT_SPACE.format(FAIL))


def copy_to_usb_drive_and_back(filepath):
    # create the mount directory
    print('USB Flash-Drive')

    if _find_storage_device():
        # remove double-quotes
        filepath = filepath.replace('"', '')
        # move the file to and from it
        name = filepath.split('/')[-1]
        try:
            run_quiet_process(
                'mv "{0}" "/mnt/usbdrive/{1}"'.format(filepath, name))
        except Exception:
            print(RESULT_SPACE.format(FAIL))
            return
        if os.path.exists('/mnt/usbdrive/{}'.format(name)):
            try:
                run_quiet_process(
                    'mv "/mnt/usbdrive/{0}" "{1}"'.format(name, filepath))
            except Exception:
                print(RESULT_SPACE.format(FAIL))
                return
            if os.path.exists(filepath):
                print(RESULT_SPACE.format(PASS))
            else:
                print(RESULT_SPACE.format(FAIL))
        else:
            print(RESULT_SPACE.format(FAIL))


def start_server(folder, filepath):
    print('\nOPEN\t--> {0}:8000/{1}\n'.format(
        _this_wifi_ip_address(),
        filepath.split('/')[-1]
    ))
    print('\nQuit\t--> CTRL-C\n\n\n')
    try:
        run_quiet_process('cd {} && python -m http.server'.format(folder))
    except KeyboardInterrupt:
        print()
        pass


if __name__ == '__main__':
    parser = args_handler.root_argparser(
        "run a suite of tests for the OT-2's hardware")
    args = parser.parse_args()
    _, driver = args_handler.build_driver(args.port)
    # put quotes around filepaths to allow whitespaces
    logging.basicConfig(filename='factory-test.log')
    data_folder_quoted = '{}'.format(DATA_FOLDER)
    video_filepath_quoted = '{}'.format(VIDEO_FILEPATH)
    atexit.register(functools.partial(_reset_lights, driver))
    atexit.register(_erase_data, video_filepath_quoted)
    _reset_lights(driver)
    _erase_data(video_filepath_quoted)
    test_smoothie_gpio(driver)
    test_switches_and_lights(driver)
    test_speaker()
    record_camera(video_filepath_quoted)
    copy_to_usb_drive_and_back(video_filepath_quoted)
    start_server(data_folder_quoted, VIDEO_FILEPATH)
    exit()



import functools
import inspect
from typing import Any, Callable, cast, Dict, Tuple, Mapping, Generic, TypeVar
from opentrons.broker import Broker
from . import types as command_types


CacheT = TypeVar("CacheT")


class InspectMemoizer(Generic[CacheT]):
    def __init__(self, method: Callable[[Any], CacheT]) -> None:
        self._method = method
        self._cache: Dict[Callable, CacheT] = {}

    def get(self, f: Callable) -> CacheT:
        v = self._cache.get(f)
        if not v:
            v = self._method(f)
            self._cache[f] = v
        return v


signature_cache = InspectMemoizer(inspect.signature)
"""Cache inspect.signature method calls"""
getfullargspec_cache = InspectMemoizer(inspect.getfullargspec)
"""Cache inspect.getfullargspec method calls"""


class CommandPublisher:
    def __init__(self, broker: Broker) -> None:
        self._broker = broker or Broker()

    @property
    def broker(self) -> Broker:
        return self._broker

    @broker.setter
    def broker(self, broker: Broker) -> None:
        self._broker = broker


CmdFunction = Callable[..., command_types.Command]


def do_publish(
        broker: Broker,
        cmd: CmdFunction,
        f: Callable,
        when: command_types.MessageSequenceId,
        res: Any,
        meta: Any,
        *args: Any, **kwargs: Any) -> None:
    """ Implement the publish so it can be called outside the decorator """
    publish_command = functools.partial(
        broker.publish,
        topic=command_types.COMMAND)
    call_args = _get_args(f, args, kwargs)
    if when == 'before':
        broker.logger.info("{}: {}".format(
            f.__qualname__,
            {k: v for k, v in call_args.items() if str(k) != 'self'}))
    getfullargspec = getfullargspec_cache.get(cmd)
    command_args = dict(
        zip(
            reversed(getfullargspec.args),
            reversed(getfullargspec.defaults
                     or [])))

    # TODO (artyom, 20170927): we are doing this to be able to use
    # the decorator in Instrument class methods, in which case
    # self is effectively an instrument.
    # To narrow the scope of this hack, we are checking if the
    # command is expecting instrument first.
    if 'instrument' in getfullargspec.args:
        # We are also checking if call arguments have 'self' and
        # don't have instruments specified, in which case
        # instruments should take precedence.
        if 'instrument' not in call_args and 'self' in call_args:
            call_args['instrument'] = call_args['self']

    command_args.update({
        key: call_args[key]
        for key in
        (set(getfullargspec.args)
         & call_args.keys())
    })

    if meta:
        command_args['meta'] = meta

    payload = cmd(**command_args)

    publish_command(
        message={**payload, '$': when})


def publish_paired(
        broker: Broker,
        cmd: CmdFunction,
        when: command_types.MessageSequenceId,
        res: Any,
        *args: Any,
        pub_type: str = 'Paired Pipettes') -> None:
    """ Implement a second publisher outside of the decorator that
    relies on the method providing all of the arguments required
    rather than binding defaults to the signature"""
    publish_command = functools.partial(
        broker.publish,
        topic=command_types.COMMAND)

    payload = cmd(*args, pub_type)

    publish_command(message={**payload, '$': when})


def _publish_dec(
        before: bool,
        after: bool,
        command: CmdFunction,
        meta: Any = None):  # noqa: ANN202
    def _decorator(f: Callable):  # noqa: ANN202
        @functools.wraps(
            f,
            updated=functools.WRAPPER_UPDATES + ('__globals__',))  # type: ignore
        def _decorated(*args, **kwargs):  # noqa: ANN202,ANN003,ANN002
            try:
                broker = cast(Broker, args[0].broker)
            except AttributeError:
                raise RuntimeError("Only methods of CommandPublisher \
                    classes should be decorated.")
            if before:
                do_publish(
                    broker, command, f, 'before', None, meta, *args, **kwargs)
            res = f(*args, **kwargs)
            if after:
                do_publish(
                    broker, command, f, 'after', res, meta, *args, **kwargs)
            return res
        return _decorated

    return _decorator


class publish:
    """ Class that allows namespaced decorators with valid mypy types

    These were previously defined by adding them as attributes to the
    publish function, which is not currently supported by mypy:
    https://github.com/python/mypy/issues/2087

    Making a simple class with these as attributes does the same thing
    but in a way that mypy can actually verify.

    Only commands inheriting from class CommandPublisher should contain
    decorators.
    """
    before = functools.partial(_publish_dec, before=True, after=False)
    after = functools.partial(_publish_dec, before=False, after=True)
    both = functools.partial(_publish_dec, before=True, after=True)


def _get_args(
        f: Callable,
        args: Tuple,
        kwargs: Mapping[str, Any]) -> Dict[str, Any]:
    # Create the initial dictionary with args that have defaults
    res = {}
    sig = signature_cache.get(f)
    ismethod = inspect.ismethod(f)
    if ismethod and args[0] is f.__self__:  # type: ignore
        args = args[1:]
    if ismethod:
        res['self'] = f.__self__  # type: ignore

    bound = sig.bind(*args, **kwargs)
    bound.apply_defaults()
    res.update(bound.arguments)
    return res



"""Base command data model and type definitions."""
from __future__ import annotations
from abc import ABC, abstractmethod
from dataclasses import dataclass
from datetime import datetime
from pydantic import BaseModel
from typing import TYPE_CHECKING, Generic, TypeVar

from ..errors import ProtocolEngineError

# convenience type alias to work around type-only circular dependency
if TYPE_CHECKING:
    from ..execution import CommandHandlers
else:
    CommandHandlers = None

ReqT = TypeVar("ReqT", bound=BaseModel)
ResT = TypeVar("ResT", bound=BaseModel)


@dataclass(frozen=True)
class CompletedCommand(Generic[ReqT, ResT]):
    """A command that has been successfully executed."""

    created_at: datetime
    started_at: datetime
    completed_at: datetime
    request: ReqT
    result: ResT


@dataclass(frozen=True)
class FailedCommand(Generic[ReqT]):
    """A command that was executed but failed."""

    created_at: datetime
    started_at: datetime
    failed_at: datetime
    request: ReqT
    error: ProtocolEngineError


@dataclass(frozen=True)
class RunningCommand(Generic[ReqT, ResT]):
    """A command that is currently being executed."""

    created_at: datetime
    started_at: datetime
    request: ReqT

    def to_completed(
        self,
        result: ResT,
        completed_at: datetime,
    ) -> CompletedCommand[ReqT, ResT]:
        """Create a CompletedCommand from a RunningCommand."""
        return CompletedCommand(
            created_at=self.created_at,
            started_at=self.started_at,
            request=self.request,
            completed_at=completed_at,
            result=result
        )

    def to_failed(
        self,
        error: ProtocolEngineError,
        failed_at: datetime,
    ) -> FailedCommand[ReqT]:
        """Create a FailedCommand from a RunningCommand."""
        return FailedCommand(
            created_at=self.created_at,
            started_at=self.started_at,
            request=self.request,
            failed_at=failed_at,
            error=error
        )


@dataclass(frozen=True)
class PendingCommand(Generic[ReqT, ResT]):
    """A command that has not yet been started."""

    created_at: datetime
    request: ReqT

    def to_running(self, started_at: datetime) -> RunningCommand[ReqT, ResT]:
        """Create a RunningCommand from a PendingCommand."""
        return RunningCommand(
            created_at=self.created_at,
            request=self.request,
            started_at=started_at,
        )


class CommandImplementation(ABC, Generic[ReqT, ResT]):
    """Abstract command implementation.

    A given command request should map to a specific command implementation,
    which defines how to:

    - Create a command resource from the request model
    - Execute the command, mapping data from execution into the result model
    """

    _request: ReqT

    def __init__(self, request: ReqT) -> None:
        """Initialize a command implementation from a command request."""
        self._request = request

    def create_command(self, created_at: datetime) -> PendingCommand[ReqT, ResT]:
        """Create a new command resource from the implementation's request."""
        return PendingCommand(request=self._request, created_at=created_at)

    @abstractmethod
    async def execute(self, handlers: CommandHandlers) -> ResT:
        """Execute the command, mapping data from execution into a response model."""
        ...



"""Resources used by command execution handlers."""
from .resource_providers import ResourceProviders
from .id_generator import IdGenerator
from .deck_data_provider import DeckDataProvider, DeckFixedLabware
from .labware_data_provider import LabwareDataProvider


__all__ = [
    "ResourceProviders",
    "IdGenerator",
    "LabwareDataProvider",
    "DeckDataProvider",
    "DeckFixedLabware",
]



"""Protocol engine command models.

The `protocol_engine.commands` module contains the data models of all
command requests and results the engine is able to handle. It also contains
models and type definitions for command resources, where a command:

- May be pending, running, completed, or failed
- Contains request and, if applicable, result data
- Has timestamps for when it was created, started, completed, and/or failed

Request and result models in the module are defined using Pydantic, because
they are part of the public input / output of the engine, and need validation
and/or schema generation.
"""

from typing import Union

from .command import PendingCommand, RunningCommand, CompletedCommand, FailedCommand

from .load_labware import LoadLabwareRequest, LoadLabwareResult
from .add_labware_definition import (
    AddLabwareDefinitionRequest, AddLabwareDefinitionResult
)
from .load_pipette import LoadPipetteRequest, LoadPipetteResult
from .move_to_well import MoveToWellRequest, MoveToWellResult
from .pick_up_tip import PickUpTipRequest, PickUpTipResult
from .drop_tip import DropTipRequest, DropTipResult
from .aspirate import AspirateRequest, AspirateResult
from .dispense import DispenseRequest, DispenseResult

CommandRequestType = Union[
    LoadLabwareRequest,
    AddLabwareDefinitionRequest,
    LoadPipetteRequest,
    MoveToWellRequest,
    PickUpTipRequest,
    DropTipRequest,
    AspirateRequest,
    DispenseRequest,
]

CommandResultType = Union[
    LoadLabwareResult,
    AddLabwareDefinitionResult,
    LoadPipetteResult,
    MoveToWellResult,
    PickUpTipResult,
    DropTipResult,
    AspirateResult,
    DispenseResult,
]

PendingCommandType = Union[
    PendingCommand[LoadLabwareRequest, LoadLabwareResult],
    PendingCommand[AddLabwareDefinitionRequest, AddLabwareDefinitionResult],
    PendingCommand[LoadPipetteRequest, LoadPipetteResult],
    PendingCommand[MoveToWellRequest, MoveToWellResult],
    PendingCommand[PickUpTipRequest, PickUpTipResult],
    PendingCommand[DropTipRequest, DropTipResult],
    PendingCommand[AspirateRequest, AspirateResult],
    PendingCommand[DispenseRequest, DispenseResult],
]

RunningCommandType = Union[
    RunningCommand[LoadLabwareRequest, LoadLabwareResult],
    RunningCommand[AddLabwareDefinitionRequest, AddLabwareDefinitionResult],
    RunningCommand[LoadPipetteRequest, LoadPipetteResult],
    RunningCommand[MoveToWellRequest, MoveToWellResult],
    RunningCommand[PickUpTipRequest, PickUpTipResult],
    RunningCommand[DropTipRequest, DropTipResult],
    RunningCommand[AspirateRequest, AspirateResult],
    RunningCommand[DispenseRequest, DispenseResult],
]

CompletedCommandType = Union[
    CompletedCommand[LoadLabwareRequest, LoadLabwareResult],
    CompletedCommand[AddLabwareDefinitionRequest, AddLabwareDefinitionResult],
    CompletedCommand[LoadPipetteRequest, LoadPipetteResult],
    CompletedCommand[MoveToWellRequest, MoveToWellResult],
    CompletedCommand[PickUpTipRequest, PickUpTipResult],
    CompletedCommand[DropTipRequest, DropTipResult],
    CompletedCommand[AspirateRequest, AspirateResult],
    CompletedCommand[DispenseRequest, DispenseResult],
]

FailedCommandType = Union[
    FailedCommand[LoadLabwareRequest],
    FailedCommand[AddLabwareDefinitionRequest],
    FailedCommand[LoadPipetteRequest],
    FailedCommand[MoveToWellRequest],
    FailedCommand[PickUpTipRequest],
    FailedCommand[DropTipRequest],
    FailedCommand[AspirateRequest],
    FailedCommand[DispenseRequest],
]

CommandType = Union[
    PendingCommandType,
    RunningCommandType,
    CompletedCommandType,
    FailedCommandType,
]

__all__ = [
    # command lifecycle state models
    "PendingCommand",
    "RunningCommand",
    "CompletedCommand",
    "FailedCommand",

    # type unions
    "CommandRequestType",
    "CommandResultType",
    "PendingCommandType",
    "RunningCommandType",
    "CompletedCommandType",
    "FailedCommandType",
    "CommandType",

    # equipment request/result models
    "LoadLabwareRequest",
    "LoadLabwareResult",
    "AddLabwareDefinitionRequest",
    "AddLabwareDefinitionResult",
    "LoadPipetteRequest",
    "LoadPipetteResult",

    # pipetting request/result models
    "MoveToWellRequest",
    "MoveToWellResult",
    "PickUpTipRequest",
    "PickUpTipResult",
    "DropTipRequest",
    "DropTipResult",
    "AspirateRequest",
    "AspirateResult",
    "DispenseRequest",
    "DispenseResult",
]



from __future__ import annotations

import re

from opentrons.protocols.geometry.well_geometry import WellGeometry
from opentrons_shared_data.labware.constants import WELL_NAME_PATTERN


class WellImplementation:

    pattern = re.compile(WELL_NAME_PATTERN, re.X)

    def __init__(self,
                 well_geometry: WellGeometry,
                 display_name: str,
                 has_tip: bool,
                 name: str):
        """
        Construct a well

        :param well_geometry: The well's geometry
        :param display_name: a string that identifies a well. Used primarily
            for debug and test purposes. Should be unique and human-readable--
            something like "Tip C3 of Opentrons 300ul Tiprack on Slot 5" or
            "Well D1 of Biorad 96 PCR Plate on Magnetic Module in Slot 1".
            This is created by the caller and passed in, so here it is just
            saved and made available.
        :param has_tip: whether a tip is present
        :param name: The well name (ie. A1, B4, C2, etc.)
        """
        self._display_name = display_name
        self._has_tip = has_tip
        self._name = name

        match = WellImplementation.pattern.match(name)
        assert match, f"could not match '{name}' using " \
                      f"pattern '{WellImplementation.pattern.pattern}'"
        self._row_name = match.group(1)
        self._column_name = match.group(2)
        self._geometry = well_geometry

    def has_tip(self) -> bool:
        return self._has_tip

    def set_has_tip(self, value: bool) -> None:
        self._has_tip = value

    def get_display_name(self) -> str:
        return self._display_name

    def get_name(self) -> str:
        """The name of the well (ie A1, A2,... B3,...C10"""
        return self._name

    def get_column_name(self) -> str:
        """The column portion of the well name"""
        return self._column_name

    def get_row_name(self) -> str:
        """The row portion of the well name"""
        return self._row_name

    def get_geometry(self) -> WellGeometry:
        return self._geometry

    def __repr__(self):
        return self.get_display_name()

    def __eq__(self, other) -> bool:
        """Assume that if name is the same then it's the same well"""
        return \
            isinstance(other, WellImplementation) and\
            self.get_name() == other.get_name()



"""JSON file reading."""

from opentrons.protocols.models import JsonProtocol
from .protocol_file import ProtocolFile


class JsonFileReader:
    """Reads and parses JSON protocol files."""

    @staticmethod
    def read(file: ProtocolFile) -> JsonProtocol:
        """Read and parse file into a JsonProtocol model."""
        contents = file.file_path.read_text(encoding="utf-8")
        return JsonProtocol.parse_raw(contents)



from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Union, Tuple, List, Optional, Dict


from opentrons.protocols.api_support.labware_like import LabwareLike
from opentrons import types, hardware_control as hc
from opentrons.commands import paired_commands as cmds
from opentrons.commands.publisher import (
    CommandPublisher, publish_paired, publish)
from opentrons.protocols.api_support.types import APIVersion
from opentrons.protocols.context.protocol_api.paired_instrument import\
    PairedInstrument

from opentrons.protocols.api_support.util import (
    requires_version, labware_column_shift, Clearances, clamp_value)
from .labware import (
    Labware, Well, OutOfTipsError, select_tiprack_from_list_paired_pipettes,
    filter_tipracks_to_start)

if TYPE_CHECKING:
    from .protocol_context import ProtocolContext
    from .instrument_context import InstrumentContext
    from opentrons.hardware_control import types as hc_types
    from opentrons.protocols.api_support.util import HardwareManager


SBS_96_WELL_SPACING = 4
SBS_384_WELL_SPACING = 7
TUBERACK_24_WELL_SPACING = 2


class UnsupportedInstrumentPairingError(Exception):
    pass


class PipettePairPickUpTipError(Exception):
    pass


class PairedInstrumentContext(CommandPublisher):

    def __init__(self,
                 primary_instrument: InstrumentContext,
                 secondary_instrument: InstrumentContext,
                 ctx: ProtocolContext,
                 pair_policy: hc_types.PipettePair,
                 api_version: APIVersion,
                 hardware_manager: HardwareManager,
                 trash: Labware,
                 log_parent: logging.Logger) -> None:
        super().__init__(ctx.broker)
        self._instruments = {
            pair_policy.primary: primary_instrument,
            pair_policy.secondary: secondary_instrument}
        self._api_version = api_version
        self._log = log_parent.getChild(repr(self))
        self._tip_racks = list(
            set(primary_instrument.tip_racks) &
            set(secondary_instrument.tip_racks))
        self._pair_policy = pair_policy

        self._starting_tip: Union[Well, None] = None
        self._last_tip_picked_up_from: Union[Well, None] = None

        self._well_bottom_clearance = Clearances(
            default_aspirate=1.0, default_dispense=1.0)

        self.trash_container = trash
        self.paired_instrument_obj = PairedInstrument(
            primary_instrument, secondary_instrument, pair_policy,
            ctx, hardware_manager, self._log)

    @property  # type: ignore
    @requires_version(2, 7)
    def api_version(self) -> APIVersion:
        return self._api_version

    @property  # type: ignore
    @requires_version(2, 7)
    def tip_racks(self) -> List[Labware]:
        return self._tip_racks

    @property  # type: ignore
    @requires_version(2, 7)
    def starting_tip(self) -> Union[Well, None]:
        """ The starting tip from which the pipette pick up
        """
        return self._starting_tip

    @starting_tip.setter
    def starting_tip(self, location: Union[Well, None]):
        self._starting_tip = location

    @requires_version(2, 7)
    def reset_tipracks(self):
        """ Reload all tips in each tip rack and reset starting tip
        """
        for tiprack in self.tip_racks:
            tiprack.reset()
        self.starting_tip = None

    @property  # type: ignore
    @requires_version(2, 7)
    def well_bottom_clearance(self) -> Clearances:
        """ The distance above the bottom of a well to aspirate or dispense.

        This is an object with attributes ``aspirate`` and ``dispense``,
        describing the default heights of the corresponding operation. The
        default is 1.0mm for both aspirate and dispense.

        When :py:meth:`aspirate` or :py:meth:`dispense` is given a
        :py:class:`.Well` rather than a full :py:class:`.Location`, the robot
        will move this distance above the bottom of the well to aspirate or
        dispense.

        Since the :py:class:`PairedInstrumentContext` controls both pipettes,
        setting this value will override any pre-set value you may have
        included on each individual :py:class:`InstrumentContext`. You
        cannot specify a different clearance value for an individual
        pipette using :py:class:`PairedInstrumentContext` at this time.

        To change, set the corresponding attribute. For instance,

        .. code-block:: python

            instr.well_bottom_clearance.aspirate = 1

        """
        return self._well_bottom_clearance

    @requires_version(2, 7)
    def pick_up_tip(
            self, location: Union[types.Location, Well] = None,
            presses: Optional[int] = None,
            increment: Optional[float] = None) -> PairedInstrumentContext:
        """
        Pick up a tip for the pipette to run liquid-handling commands with
        a paired pipette object.

        If no location is passed, the Pipette will pick up the next available
        tip in its :py:attr:`PairedInstrumentContext.tip_racks` list. Pick up
        tip will determine the next available tip based on both the primary
        and secondary target.

        If you choose to pass in a tip location, you must pass in the
        tiprack location that you wish to move based on the primary pipette.

        The tip to pick up can be manually specified with the ``location``
        argument. The ``location`` argument can be specified in several ways:

        * If the only thing to specify is which well from which to pick
          up a tip, `location` can be a :py:class:`.Well`. For instance,
          if you have a tip rack in a variable called `tiprack`, you can
          pick up a specific tip from it with
          ``instr.pick_up_tip(tiprack.wells()[0])``. This style of call can
          be used to make the robot pick up a tip from a tip rack that
          was not specified when creating the
          :py:class:`.PairedInstrumentContext`.

        * If the position to move to in the well needs to be specified,
          for instance to tell the robot to run its pick up tip routine
          starting closer to or farther from the top of the tip,
          `location` can be a :py:class:`.types.Location`; for instance,
          you can call ``instr.pick_up_tip(tiprack.wells()[0].top())``.

        :param location: The location from which to pick up a tip.
        :type location: :py:class:`.types.Location` or :py:class:`.Well` to
                        pick up a tip from.
        :param presses: The number of times to lower and then raise the pipette
                        when picking up a tip, to ensure a good seal (0 [zero]
                        will result in the pipette hovering over the tip but
                        not picking it up--generally not desireable, but could
                        be used for dry-run).
        :type presses: int
        :param increment: The additional distance to travel on each successive
                          press (e.g.: if `presses=3` and `increment=1.0`, then
                          the first press will travel down into the tip by
                          3.5mm, the second by 4.5mm, and the third by 5.5mm).
        :type increment: float

        :returns: This instance
        :raises OutOfTipsError: if there are no more tips to pick up from a
        tiprack
        :raises PipettePairPickUpTipError: If you specify a location that both
        pipettes cannot pick up a tip from, for example, the last column
        of the tiprack

        .. note::

            The current spacing of the two pipettes is approximately 34
            millimeters, or 5 columns in a 96 well SBS footprint.
            In practice, this means that you can only pick up from 10
            columns of the tiprack rather than the full 12 columns of a 96
            SBS footprint when utilizing the
            :py:attr:`PairedInstrumentContext.pick_up_tip` command.
            You can, however, use the regular
            :py:attr:`InstrumentContext.pick_up_tip` to pick up the
            remaining tips from the tiprack. See the code snippet below.

        .. code-block:: python

            >>> tiprack = ctx.load_labware('opentrons_96_tiprack_300ul', 1)
            >>> left = ctx.load_instrument('p300_multi', 'left', tip_racks=[tiprack])
            >>> right = ctx.load_instrument('p300_multi', 'right', tip_racks=[tiprack])
            >>> paired = right.pair_with(left)
            >>> paired.pick_up_tip() # Pick up tips from the first and fifth column
            >>> paired.drop_tip()
            >>> left.pick_up_tip() # Pick up tips from the second column
            >>> left.drop_tip()
            >>> right.pick_up_tip() # Pick up tips from third column
            >>> right.drop_tip()
        """
        if location and isinstance(location, types.Location):
            if location.labware.is_labware:
                tiprack = location.labware.as_labware()
                primary_channels =\
                    self._instruments[self._pair_policy.primary].channels
                target: Well =\
                    tiprack.next_tip(primary_channels)  # type: ignore
                if not target:
                    raise OutOfTipsError
            elif location.labware.is_well:
                tiprack = location.labware.parent.as_labware()
                target = location.labware.as_well()
            else:
                raise TypeError(
                    "The location must be in a well or labware."
                )
        elif location and isinstance(location, Well):
            tiprack = location.parent
            target = location
        elif not location:
            tiprack, target = self._next_available_tip()
        else:
            raise TypeError(
                "If specified, location should be an instance of "
                "types.Location (e.g. the return value from "
                "tiprack.wells()[0].top()) or a Well (e.g. tiprack.wells()[0]."
                " However, it is a {}".format(location))
        assert tiprack.is_tiprack, "{} is not a tiprack".format(str(tiprack))

        # We must check that you can also pick up from
        # that tiprack with the second pipette attached.
        secondary_target = self._get_secondary_target(tiprack, target)
        primary_pipette = self._instruments[self._pair_policy.primary]
        tip_length = primary_pipette._tip_length_for(tiprack)

        instruments = list(self._instruments.values())
        targets = [target, secondary_target]
        publish_paired(self.broker, cmds.paired_pick_up_tip,
                       'before', None, instruments, targets)
        self.paired_instrument_obj.pick_up_tip(
            target, secondary_target, tiprack, presses, increment, tip_length)
        self._last_tip_picked_up_from = target
        publish_paired(self.broker, cmds.paired_pick_up_tip,
                       'after', self, instruments, targets)
        return self

    @requires_version(2, 7)
    def drop_tip(
            self,
            location: Union[types.Location, Well] = None,
            home_after: bool = True)\
            -> PairedInstrumentContext:
        """
        Drop the current tip.

        If no location is passed, both pipettes will drop their
        tips into :py:attr:`trash_container`, which if not specified defaults
        to the fixed trash in slot 12.

        If you pass in a location, you must pass in the tiprack location that
        you wish to move based on the primary pipette.

        The location in which to drop the tip can be manually specified with
        the ``location`` argument. The ``location`` argument can be specified
        several ways:

            - If the only thing to specify is which well into which to drop
              a tip, `location` can be a :py:class:`.Well`. For instance,
              if you have a tip rack in a variable called `tiprack`, you can
              drop a tip into a specific well on that tiprack with the call
              `instr.drop_tip(tiprack.wells()[0])`. This style of call can
              be used to make the robot drop a tip into arbitrary labware.
            - If the position to drop the tip from as well as the
              :py:class:`.Well` to drop the tip into needs to be specified,
              for instance to tell the robot to drop a tip from an unusually
              large height above the tiprack, `location`
              can be a :py:class:`.types.Location`; for instance, you can call
              `instr.drop_tip(tiprack.wells()[0].top())`.

        :param location: The location to drop the tip
        :type location: :py:class:`.types.Location` or :py:class:`.Well` or
                        None
        :param home_after: Whether to home the plunger after dropping the tip
                           (defaults to ``True``). The plungeer must home after
                           dropping tips because the ejector shroud that pops
                           the tip off the end of the pipette is driven by the
                           plunger motor, and may skip steps when dropping the
                           tip.

        :returns: This instance

        :raises TyepError: If a location is not relative to a well.
        """
        is_trash = False
        tiprack = None
        if location and isinstance(location, types.Location):
            if location.labware.is_well:
                target = location
                tiprack = location.labware.as_well().parent
            else:
                raise TypeError(
                    "If a location is specified as a types.Location (for "
                    "instance, as the result of a call to "
                    "tiprack.wells()[0].top()) it must be a location "
                    "relative to a well, since that is where a tip is "
                    "dropped. The passed location, however, is in "
                    "reference to {}".format(location.labware))
        elif location and isinstance(location, Well):
            if LabwareLike(location).is_fixed_trash():
                target = location.top()
            else:
                primary_pipette = self._instruments[self._pair_policy.primary]
                target = self._drop_tip_target(location, primary_pipette)
                tiprack = location.parent
        elif not location:
            target = self.trash_container.wells()[0].top()
            is_trash = True
        else:
            raise TypeError(
                "If specified, location should be an instance of "
                "types.Location (e.g. the return value from "
                "tiprack.wells()[0].top()) or a Well (e.g. tiprack.wells()[0]."
                " However, it is a {}".format(location))

        instruments = list(self._instruments.values())
        if not is_trash and tiprack and isinstance(target.labware, Well):
            targets = [
                target,
                self._get_secondary_target(tiprack, target.labware.as_well())]
        else:
            targets = [target]
        publish_paired(self.broker, cmds.paired_drop_tip,
                       'before', None, instruments, targets)
        self.paired_instrument_obj.drop_tip(target, home_after)
        publish_paired(self.broker, cmds.paired_drop_tip,
                       'after', self, instruments, targets)
        self._last_tip_picked_up_from = None
        return self

    @requires_version(2, 7)
    def aspirate(self,
                 volume: Optional[float] = None,
                 location: Union[types.Location, Well] = None,
                 rate: float = 1.0) -> PairedInstrumentContext:
        """
        Aspirate a volume of liquid (in microliters/uL) using both pipettes
        from a specified location. You must pass in the location you
        wish the primary pipette to move to.

        For example, if your primary pipette is the left pipette and you
        wish for it to move to well ``A1`` of a 96 SBS format plate you should
        pass in well ``A1`` of this plate. The right pipette will automatically
        aspirate from well ``A5`` of this plate.

        If only a volume is passed, the pipettes will aspirate
        from their current position. If only a location is passed (as in
        ``instr.aspirate(location=wellplate['A1'])``,
        :py:meth:`aspirate` will default to the amount of volume
        available in both pipettes being used.

        :param volume: The volume to aspirate, in microliters. If not
                       specified, :py:attr:`max_volume`.
        :type volume: int or float
        :param location: Where to aspirate from. If `location` is a
                         :py:class:`.Well`, the robot will aspirate from
                         :py:obj:`well_bottom_clearance.aspirate` mm
                         above the bottom of the well. If `location` is a
                         :py:class:`.Location` (i.e. the result of
                         :py:meth:`.Well.top` or :py:meth:`.Well.bottom`), the
                         robot will aspirate from the exact specified location.
                         If unspecified, the robot will aspirate from the
                         current position.
        :param rate: The relative plunger speed for this aspirate. During
                     this aspirate, the speed of the plunger will be
                     `rate` * :py:attr:`aspirate_speed`. If not specified,
                     defaults to 1.0 (speed will not be modified).
        :type rate: float
        :returns: This instance.

        .. note::

            If ``aspirate`` is called with a single argument, it will not try
            to guess whether the argument is a volume or location - it is
            required to be a volume. If you want to call ``aspirate`` with only
            a location, specify it as a keyword argument:
            ``instr.aspirate(location=wellplate['A1'])``

        """
        self._log.debug("aspirate {} from {} at {}"
                        .format(volume,
                                location if location else 'current position',
                                rate))

        loc: Optional[types.Location] = None
        if isinstance(location, Well):
            point, well = location.bottom()
            loc = types.Location(
                point + types.Point(0, 0,
                                    self.well_bottom_clearance.aspirate),
                well)
        elif location is not None and not isinstance(location, types.Location):
            raise TypeError(
                'location should be a Well or Location, but it is {}'
                .format(location))
        else:
            loc = location

        if not volume:
            c_vol = self._get_minimum_available_volume(self._instruments)
        else:
            c_vol = volume

        instruments = list(self._instruments.values())
        primary_loc, aspirate_func =\
            self.paired_instrument_obj.aspirate(volume, loc, rate)
        if primary_loc.labware.is_well:
            well = primary_loc.labware.as_well()
            labware = well.parent
            locations = [
                primary_loc,
                self._get_secondary_target(labware, well)]
        else:
            locations = [primary_loc]
        publish_paired(self.broker, cmds.paired_aspirate, 'before',
                       None, instruments, c_vol, locations, rate)
        aspirate_func()
        publish_paired(self.broker, cmds.paired_aspirate, 'after',
                       self, instruments, c_vol, locations, rate)
        return self

    @requires_version(2, 7)
    def dispense(self,
                 volume: Optional[float] = None,
                 location: Union[types.Location, Well] = None,
                 rate: float = 1.0) -> PairedInstrumentContext:
        """
        Dispense a volume of liquid (in microliters/uL) using both pipettes
        into the specified location.  You must pass in the location you
        wish the primary pipette to move to.

        For example, if your primary pipette is the left pipette and you
        wish for it to move to well ``A1`` of a 96 SBS format plate you should
        pass in well ``A1`` of this plate. The right pipette will automatically
        dispense into well ``A5`` of this plate.

        If only a volume is passed, the pipette will dispense from its current
        position. If only a location is passed (as in
        ``instr.dispense(location=wellplate['A1'])``), all of the liquid
        aspirated into both pipettes will be dispensed (this volume is
        accessible through :py:attr:`current_volume`).

        :param volume: The volume of liquid to dispense, in microliters. If not
                       specified, defaults to :py:attr:`current_volume`.
        :type volume: int or float

        :param location: Where to dispense into. If `location` is a
                         :py:class:`.Well`, the robot will dispense into
                         :py:obj:`well_bottom_clearance.dispense` mm
                         above the bottom of the well. If `location` is a
                         :py:class:`.Location` (i.e. the result of
                         :py:meth:`.Well.top` or :py:meth:`.Well.bottom`), the
                         robot will dispense into the exact specified location.
                         If unspecified, the robot will dispense into the
                         current position.
        :param rate: The relative plunger speed for this dispense. During
                     this dispense, the speed of the plunger will be
                     `rate` * :py:attr:`dispense_speed`. If not specified,
                     defaults to 1.0 (speed will not be modified).
        :type rate: float

        :returns: This instance.

        .. note::

            If ``dispense`` is called with a single argument, it will not try
            to guess whether the argument is a volume or location - it is
            required to be a volume. If you want to call ``dispense`` with only
            a location, specify it as a keyword argument:
            ``instr.dispense(location=wellplate['A1'])``

        """
        self._log.debug("dispense {} from {} at {}"
                        .format(volume,
                                location if location else 'current position',
                                rate))
        if isinstance(location, Well):
            if LabwareLike(location).is_fixed_trash():
                loc = location.top()
            else:
                point, well = location.bottom()
                loc = types.Location(
                    point + types.Point(0, 0,
                                        self.well_bottom_clearance.dispense),
                    well)
        elif location is not None and not isinstance(location, types.Location):
            raise TypeError(
                'location should be a Well or Location, but it is {}'
                .format(location))
        else:
            loc = location

        if not volume:
            c_vol = self._get_minimum_current_volume(self._instruments)
        else:
            c_vol = volume

        instruments = list(self._instruments.values())
        primary_loc, dispense_func =\
            self.paired_instrument_obj.dispense(volume, loc, rate)

        if isinstance(primary_loc.labware, Well):
            well = primary_loc.labware.as_well()
            labware = well.parent
            locations = [
                primary_loc,
                self._get_secondary_target(labware, well)]
        else:
            locations = [primary_loc]
        publish_paired(self.broker, cmds.paired_dispense, 'before',
                       None, instruments, c_vol, locations, rate)
        dispense_func()
        publish_paired(self.broker, cmds.paired_dispense, 'after',
                       self, instruments, c_vol, locations, rate)
        return self

    @publish.both(command=cmds.air_gap)
    @requires_version(2, 7)
    def air_gap(self,
                volume: Optional[float] = None,
                height: Optional[float] = None) -> PairedInstrumentContext:
        """
        Pull air into the pipette current tip at the current location

        :param volume: The amount in uL to aspirate air into the tube.
                       (Default will use all remaining volume in tip)
        :type volume: float

        :param height: The number of millimiters to move above the current Well
                       to air-gap aspirate. (Default: 5mm above current Well)
        :type height: float

        :raises NoTipAttachedError: If no tip is attached to the pipette

        :raises RuntimeError: If location cache is None.
                              This should happen if `touch_tip` is called
                              without first calling a method that takes a
                              location (eg, :py:meth:`.aspirate`,
                              :py:meth:`dispense`)

        :returns: This instance

        .. note::

            Both ``volume`` and height are optional, but unlike previous API
            versions, if you want to specify only ``height`` you must do it
            as a keyword argument: ``pipette.air_gap(height=2)``. If you
            call ``air_gap`` with only one unnamed argument, it will always
            be interpreted as a volume.


        """
        for instr in self._instruments.values():
            if not instr.hw_pipette['has_tip']:
                raise hc.NoTipAttachedError(
                    'Pipette has no tip. Aborting air_gap')

        if height is None:
            height = 5
        self.paired_instrument_obj.air_gap(volume, height)
        return self

    @requires_version(2, 7)
    def blow_out(self,
                 location: Union[types.Location, Well] = None
                 ) -> PairedInstrumentContext:
        """
        Blow liquid out of the tip.

        If :py:attr:`dispense` is used to completely empty a pipette,
        usually a small amount of liquid will remain in the tip. This
        method moves the plunger past its usual stops to fully remove
        any remaining liquid from the tip. Regardless of how much liquid
        was in the tip when this function is called, after it is done
        the tip will be empty.

        :param location: The location to blow out into. If not specified,
                         defaults to the current location of the pipette
        :type location: :py:class:`.Well` or :py:class:`.Location` or None

        :raises RuntimeError: If no location is specified and location cache is
                              None. This should happen if `blow_out` is called
                              without first calling a method that takes a
                              location (eg, :py:meth:`.aspirate`,
                              :py:meth:`dispense`)
        :returns: This instance
        """
        if isinstance(location, Well):
            if location.parent.is_tiprack:
                self._log.warning('Blow_out being performed on a tiprack. '
                                  'Please re-check your code')
            loc = location.top()
        elif location is not None and\
                not isinstance(location, types.Location):
            raise TypeError(
                'location should be a Well or Location, but it is {}'
                .format(location))
        else:
            loc = location

        locations: Optional[List] = None
        if loc and isinstance(loc.labware, Well):
            locations = self._get_locations(loc)

        instruments = list(self._instruments.values())
        publish_paired(self.broker, cmds.paired_blow_out,
                       'before', None, instruments, locations)
        self.paired_instrument_obj.blow_out(loc)
        publish_paired(self.broker, cmds.paired_blow_out,
                       'after', self, instruments, locations)
        return self

    @requires_version(2, 7)
    def mix(self,
            repetitions: int = 1,
            volume: Optional[float] = None,
            location: Union[types.Location, Well] = None,
            rate: float = 1.0) -> PairedInstrumentContext:
        """
        Mix a volume of liquid (uL) using this pipette.
        If no location is specified, the pipette will mix from its current
        position. If no volume is passed, ``mix`` will default to the
        pipette's :py:attr:`max_volume`.

        :param repetitions: how many times the pipette should mix (default: 1)
        :param volume: number of microliters to mix (default:
                       :py:attr:`max_volume`)
        :param location: a Well or a position relative to well.
                         e.g, `plate.rows()[0][0].bottom()`
        :type location: types.Location
        :param rate: Set plunger speed for this mix, where,
                     ``speed = rate * (aspirate_speed or dispense_speed)``
        :raises NoTipAttachedError: If no tip is attached to the pipette.
        :returns: This instance

        .. note::

            All the arguments to ``mix`` are optional; however, if you do
            not want to specify one of them, all arguments after that one
            should be keyword arguments. For instance, if you do not want
            to specify volume, you would call
            ``pipette.mix(1, location=wellplate['A1'])``. If you do not
            want to specify repetitions, you would call
            ``pipette.mix(volume=10, location=wellplate['A1'])``. Unlike
            previous API versions, ``mix`` will not attempt to guess your
            inputs; the first argument will always be interpreted as
            ``repetitions``, the second as ``volume``, and the third as
            ``location`` unless you use keywords.

        """
        self._log.debug(
            'mixing {}uL with {} repetitions in {} at rate={}'.format(
                volume, repetitions,
                location if location else 'current position', rate))
        for instr in self._instruments.values():
            if not instr.hw_pipette['has_tip']:
                raise hc.NoTipAttachedError(
                    'Pipette has no tip. Aborting mix()')

        if not volume:
            c_vol = self._get_minimum_available_volume(self._instruments)
        else:
            c_vol = volume

        instruments = list(self._instruments.values())
        locations: Optional[List] = None
        if location:
            locations = self._get_locations(location)
        publish_paired(self.broker, cmds.paired_mix, 'before', None,
                       instruments, locations, repetitions, c_vol)
        self.aspirate(volume, location, rate)
        while repetitions - 1 > 0:
            self.dispense(volume, rate=rate)
            self.aspirate(volume, rate=rate)
            repetitions -= 1
        self.dispense(volume, rate=rate)
        publish_paired(self.broker, cmds.paired_mix, 'after', self,
                       instruments, locations, repetitions, c_vol)
        return self

    @requires_version(2, 7)
    def touch_tip(self,
                  location: Optional[Well] = None,
                  radius: float = 1.0,
                  v_offset: float = -1.0,
                  speed: float = 60.0) -> PairedInstrumentContext:
        """
        Touch the pipette tip to the sides of a well, with the intent of
        removing left-over droplets. For :py:class:`PairedInstrumentContext`
        we will only use the primary pipette for collision detection purposes.

        :param location: If no location is passed, pipette will
                         touch tip at current well's edges
        :type location: :py:class:`.Well` or None
        :param radius: Describes the proportion of the target well's
                       radius. When `radius=1.0`, the pipette tip will move to
                       the edge of the target well; when `radius=0.5`, it will
                       move to 50% of the well's radius. Default: 1.0 (100%)
        :type radius: float
        :param v_offset: The offset in mm from the top of the well to touch tip
                         A positive offset moves the tip higher above the well,
                         while a negative offset moves it lower into the well
                         Default: -1.0 mm
        :type v_offset: float
        :param speed: The speed for touch tip motion, in mm/s.
                      Default: 60.0 mm/s, Max: 80.0 mm/s, Min: 20.0 mm/s
        :type speed: float
        :raises NoTipAttachedError: if no tip is attached to the pipette
        :raises RuntimeError: If no location is specified and location cache is
                              None. This should happen if `touch_tip` is called
                              without first calling a method that takes a
                              location (eg, :py:meth:`.aspirate`,
                              :py:meth:`dispense`)
        :returns: This instance

        .. note::

            This is behavior change from legacy API (which accepts any
            :py:class:`.Placeable` as the ``location`` parameter)

        """
        for instr in self._instruments.values():
            if not instr.hw_pipette['has_tip']:
                raise hc.NoTipAttachedError(
                    'Pipette has no tip to touch_tip()')

        checked_speed = clamp_value(speed, 80, 1, 'touch_tip:')

        instruments = list(self._instruments.values())
        locations: Optional[List] = None
        if location:
            locations = [
                location,
                self._get_secondary_target(location.parent, location)]

        publish_paired(self.broker, cmds.paired_touch_tip,
                       'before', None, instruments, locations)
        self.paired_instrument_obj.touch_tip(
            location, radius, v_offset, checked_speed)
        publish_paired(self.broker, cmds.paired_touch_tip,
                       'after', self, instruments, locations)
        return self

    @publish.both(command=cmds.return_tip)
    @requires_version(2, 7)
    def return_tip(self,
                   home_after: bool = True) -> PairedInstrumentContext:
        """
        If a tip is currently attached to both of the pipettes, then it will
        return the tip to it's location in the tiprack.

        It will not reset tip tracking so the well flag will remain False.

        :returns: This instance
        """
        for mount, instr in self._instruments.items():
            if not instr.hw_pipette['has_tip']:
                self._log.warning('Pipette has no tip to return')
        loc = self._last_tip_picked_up_from
        if not isinstance(loc, Well):
            raise TypeError('Last tip location should be a Well but it is: '
                            f'{loc}')
        primary_pipette = self._instruments[self._pair_policy.primary]
        drop_loc = self._drop_tip_target(loc, primary_pipette)
        self.drop_tip(drop_loc, home_after=home_after)
        return self

    @requires_version(2, 7)
    def move_to(self, location: types.Location, force_direct: bool = False,
                minimum_z_height: Optional[float] = None,
                speed: Optional[float] = None
                ) -> PairedInstrumentContext:
        """ Move the full gantry with the primary pipette used as the basis
        for the XY position within a given labware.

        :param location: The location to move to.
        :type location: :py:class:`.types.Location`
        :param force_direct: If set to true, move directly to destination
                        without arc motion.
        :param minimum_z_height: When specified, this Z margin is able to raise
                                 (but never lower) the mid-arc height.
        :param speed: The speed at which to move. If no speed is specified,
                      this function will utilize the primary pipette's speed.
                      This controls the straight linear speed of the motion;
                      to limit individual axis speeds, you can use
                      :py:attr:`.ProtocolContext.max_speeds`.
        """
        instruments = list(self._instruments.values())
        locations: Optional[List] = None
        if location:
            locations = self._get_locations(location)

        publish_paired(self.broker, cmds.paired_move_to,
                       'before', None, instruments, locations)
        self.paired_instrument_obj.move_to(
            location, force_direct, minimum_z_height, speed)
        publish_paired(self.broker, cmds.paired_move_to,
                       'after', None, instruments, locations)
        return self

    def _next_available_tip(self) -> Tuple[Labware, Well]:
        # Here we will find the next available tip for
        # both the primary and secondary tip with a spacing
        # of approximately 34 mm.
        start = self.starting_tip
        primary_channels =\
            self._instruments[self._pair_policy.primary].channels
        secondary_channels =\
            self._instruments[self._pair_policy.secondary].channels
        if start is None:
            return select_tiprack_from_list_paired_pipettes(
                self.tip_racks, primary_channels, secondary_channels)
        else:
            return select_tiprack_from_list_paired_pipettes(
                filter_tipracks_to_start(start, self.tip_racks),
                primary_channels, secondary_channels, starting_point=start)

    def _get_locations(self, location: Union[types.Location, Well]) -> List:
        if isinstance(location, Well):
            labware = location.parent
            well = location
            return [location, self._get_secondary_target(labware, well)]
        elif location.labware.is_well:
            well = location.labware.as_well()
            labware = well.parent
            return [location, self._get_secondary_target(labware, well)]
        return []

    @staticmethod
    def _drop_tip_target(
            location: Well,
            pipette: InstrumentContext) -> types.Location:
        tr = location.parent
        assert tr.is_tiprack
        z_height = pipette.return_height * tr.tip_length
        return location.top(-z_height)

    @staticmethod
    def _get_secondary_target(labware: Labware, primary_loc: Well) -> Well:
        try:
            if len(labware.columns()) == 24:
                spacing = SBS_384_WELL_SPACING
                modulo_value = 16
            elif len(labware.columns()) == 6:
                spacing = TUBERACK_24_WELL_SPACING
                modulo_value = 4
            else:
                spacing = SBS_96_WELL_SPACING
                modulo_value = 8
            return labware_column_shift(
                primary_loc, labware, spacing, modulo_value)
        except IndexError:
            raise PipettePairPickUpTipError(
                f'The location you specified ({primary_loc}) is'
                'incompatible with a PipettePair pickup.')

    @staticmethod
    def _get_minimum_available_volume(
            instruments: Dict[types.Mount, InstrumentContext]) -> float:
        return min(
            instr.hw_pipette['available_volume']
            for instr in instruments.values())

    @staticmethod
    def _get_minimum_current_volume(
            instruments: Dict[types.Mount, InstrumentContext]) -> float:
        return min(
            instr.hw_pipette['current_volume']
            for instr in instruments.values())



from .session import Session, SessionManager
from .routers import MainRouter
from .calibration import CalibrationManager

__all__ = [
    'Session',
    'SessionManager',
    'MainRouter',
    'CalibrationManager'
]



from typing import TYPE_CHECKING
from opentrons.protocols.execution.execute_json_v3 import (
    _blowout, _pick_up_tip, _drop_tip, _aspirate, _dispense, _touch_tip,
    _air_gap)
from opentrons.protocols.execution.execute_json_v4 import _engage_magnet, \
    _disengage_magnet, _temperature_module_set_temp, \
    _temperature_module_deactivate, \
    _temperature_module_await_temp, \
    _thermocycler_close_lid, \
    _thermocycler_open_lid, \
    _thermocycler_deactivate_block, \
    _thermocycler_deactivate_lid, \
    _thermocycler_set_block_temperature, \
    _thermocycler_set_lid_temperature, \
    _thermocycler_run_profile
from opentrons.protocols.execution.execute_json_v5 import _move_to_well

from opentrons.protocol_api.module_contexts import ThermocyclerContext


from opentrons_shared_data.protocol.constants import (
    JsonPipetteCommand, JsonMagneticModuleCommand,
    JsonTemperatureModuleCommand, JsonThermocyclerCommand
)

if TYPE_CHECKING:
    from opentrons.protocols.execution.dev_types import (
        PipetteDispatch, JsonV4MagneticModuleDispatch,
        JsonV4TemperatureModuleDispatch,
        JsonV4ThermocyclerDispatch
    )


pipette_command_map: 'PipetteDispatch' = {
    JsonPipetteCommand.blowout.value: _blowout,
    JsonPipetteCommand.pickUpTip.value: _pick_up_tip,
    JsonPipetteCommand.dropTip.value: _drop_tip,
    JsonPipetteCommand.aspirate.value: _aspirate,
    JsonPipetteCommand.dispense.value: _dispense,
    JsonPipetteCommand.touchTip.value: _touch_tip,
    JsonPipetteCommand.airGap.value: _air_gap,
    JsonPipetteCommand.moveToWell.value: _move_to_well
}

magnetic_module_command_map: 'JsonV4MagneticModuleDispatch' = {
    JsonMagneticModuleCommand.magneticModuleEngageMagnet.value: _engage_magnet,
    JsonMagneticModuleCommand.magneticModuleDisengageMagnet.value:
    _disengage_magnet,
}

temperature_module_command_map: 'JsonV4TemperatureModuleDispatch' = {
    JsonTemperatureModuleCommand.temperatureModuleSetTargetTemperature.value:
        _temperature_module_set_temp,
    JsonTemperatureModuleCommand.temperatureModuleDeactivate.value:
        _temperature_module_deactivate,
    JsonTemperatureModuleCommand.temperatureModuleAwaitTemperature.value:
        _temperature_module_await_temp
}


def tc_do_nothing(module: ThermocyclerContext, params) -> None:
    pass


thermocycler_module_command_map: 'JsonV4ThermocyclerDispatch' = {
    JsonThermocyclerCommand.thermocyclerCloseLid.value:
    _thermocycler_close_lid,
    JsonThermocyclerCommand.thermocyclerOpenLid.value: _thermocycler_open_lid,
    JsonThermocyclerCommand.thermocyclerDeactivateBlock.value:
    _thermocycler_deactivate_block,
    JsonThermocyclerCommand.thermocyclerDeactivateLid.value:
    _thermocycler_deactivate_lid,
    JsonThermocyclerCommand.thermocyclerSetTargetBlockTemperature.value:
    _thermocycler_set_block_temperature,
    JsonThermocyclerCommand.thermocyclerSetTargetLidTemperature.value:
    _thermocycler_set_lid_temperature,
    JsonThermocyclerCommand.thermocyclerRunProfile.value:
    _thermocycler_run_profile,
    # NOTE: the thermocyclerAwaitX commands are expected to always
    # follow a corresponding SetX command, which is implemented as
    # blocking. Then nothing needs to be done for awaitX commands.
    JsonThermocyclerCommand.thermocyclerAwaitBlockTemperature.value:
    tc_do_nothing,
    JsonThermocyclerCommand.thermocyclerAwaitLidTemperature.value:
    tc_do_nothing,
    JsonThermocyclerCommand.thermocyclerAwaitProfileComplete.value:
    tc_do_nothing,
}



"""Protocol runner factory."""
from typing import Optional

from opentrons.protocol_engine import ProtocolEngine
from opentrons.protocols.runner import CommandTranslator

from .abstract_file_runner import AbstractFileRunner
from .json_file_runner import JsonFileRunner
from .json_file_reader import JsonFileReader
from .command_queue_worker import CommandQueueWorker
from .protocol_file import ProtocolFileType, ProtocolFile


def create_file_runner(
    protocol_file: Optional[ProtocolFile],
    engine: ProtocolEngine,
) -> AbstractFileRunner:
    """Construct a wired-up protocol runner instance.

    Arguments:
        protocol_file: Protocol file the runner will be using. If `None`,
            returns a basic runner for ProtocolEngine usage without a file.
        engine: The protocol engine interface the runner will use.

    Returns:
        A runner appropriate for the requested protocol type.
    """
    if protocol_file is not None and protocol_file.file_type == ProtocolFileType.JSON:
        return JsonFileRunner(
            file=protocol_file,
            protocol_engine=engine,
            file_reader=JsonFileReader(),
            command_translator=CommandTranslator(),
            command_queue_worker=CommandQueueWorker(),
        )

    raise NotImplementedError("Other runner types not yet supported")



"""Add labware command."""
from __future__ import annotations
from opentrons.protocol_engine.commands.command import (
    CommandImplementation, CommandHandlers
)
from pydantic import BaseModel, Field
from opentrons.protocols.models import LabwareDefinition


class AddLabwareDefinitionRequest(BaseModel):
    """Request to add a labware definition."""
    definition: LabwareDefinition = Field(
        ...,
        description="The labware definition."
    )

    def get_implementation(self) -> AddLabwareDefinitionImplementation:
        """Get the execution implementation of the AddLabwareDefinitionRequest."""
        return AddLabwareDefinitionImplementation(self)


class AddLabwareDefinitionResult(BaseModel):
    """Result of add labware request."""

    loadName: str = Field(
        ...,
        description="Name used to reference a labware definition.",
    )
    namespace: str = Field(
        ...,
        description="The namespace the labware definition belongs to.",
    )
    version: int = Field(
        ...,
        description="The labware definition version.",
    )


class AddLabwareDefinitionImplementation(
    CommandImplementation[AddLabwareDefinitionRequest,
                          AddLabwareDefinitionResult]
):
    """Add labware command implementation."""

    async def execute(
            self,
            handlers: CommandHandlers
    ) -> AddLabwareDefinitionResult:
        """Execute the add labware definition request."""
        return AddLabwareDefinitionResult(
            loadName=self._request.definition.parameters.loadName,
            namespace=self._request.definition.namespace,
            version=self._request.definition.version
        )



"""File runner interfaces for JSON protocols."""
from opentrons.protocol_engine import ProtocolEngine
from opentrons.protocols.runner import CommandTranslator

from .abstract_file_runner import AbstractFileRunner
from .json_file_reader import JsonFileReader
from .command_queue_worker import CommandQueueWorker
from .protocol_file import ProtocolFile


class JsonFileRunner(AbstractFileRunner):
    """JSON protocol file runner."""

    def __init__(
        self,
        file: ProtocolFile,
        file_reader: JsonFileReader,
        protocol_engine: ProtocolEngine,
        command_translator: CommandTranslator,
        command_queue_worker: CommandQueueWorker,
    ) -> None:
        """JSON file runner constructor.

        Args:
            file: a JSON protocol file
            file_reader: an interface to read the file into a data model.
            protocol_engine: instance of the Protocol Engine
            command_translator: the JSON command translator
            command_queue_worker: Command Queue worker
        """
        self._file = file
        self._file_reader = file_reader
        self._protocol_engine = protocol_engine
        self._command_translator = command_translator
        self._command_queue_worker = command_queue_worker

    def load(self) -> None:
        """Translate JSON commands and send them to protocol engine."""
        protocol = self._file_reader.read(self._file)

        for json_cmd in protocol.commands:
            translated_items = self._command_translator.translate(json_cmd)
            for cmd in translated_items:
                self._protocol_engine.add_command(cmd)

    def play(self) -> None:
        """Start (or un-pause) running the JSON protocol file."""
        self._command_queue_worker.play()

    def pause(self) -> None:
        """Pause the running JSON protocol file's execution."""
        self._command_queue_worker.pause()

    def stop(self) -> None:
        """Cancel the running JSON protocol file."""
        self._command_queue_worker.stop()



from datetime import timedelta

from . import types as command_types


def comment(msg: str) -> command_types.CommentCommand:
    text = msg
    return {
        'name': command_types.COMMENT,
        'payload': {'text': text}
    }


def delay(
        seconds: float,
        minutes: float,
        msg: str = None) -> command_types.DelayCommand:
    td = timedelta(minutes=minutes, seconds=seconds)
    actual_min, actual_sec = divmod(td.total_seconds(), 60)

    text = (
        f"Delaying for {int(actual_min)} minutes and "
        f"{round(actual_sec, 3)} seconds"
    )

    if msg:
        text = f"{text}. {msg}"

    return {
        'name': command_types.DELAY,
        'payload': {
            'minutes': actual_min,
            'seconds': actual_sec,
            'text': text
        }
    }


def pause(msg: str = None) -> command_types.PauseCommand:
    text = 'Pausing robot operation'
    if msg:
        text = text + ': {}'.format(msg)
    return {
        'name': command_types.PAUSE,
        'payload': {
            'text': text,
            'userMessage': msg,
        }
    }


def resume() -> command_types.ResumeCommand:
    return {
        'name': command_types.RESUME,
        'payload': {'text': 'Resuming robot operation'}
    }



"""
hardware_control: The sole authority for controlling the hardware of an OT2.

The hardware_control module presents a unified api for the lowest level of
hardware command that takes into account the robot as a whole. For instance,
it presents an API for moving a specific pipette mount (not a specific motor
or axis)  to a deck-absolute point (not a Smoothie-coordinate point).

This module is not for use outside the opentrons api module. Higher-level
functions are available elsewhere.
"""

from .adapters import SynchronousAdapter
from .api import API, PauseManager
from .controller import Controller
from .simulator import Simulator
from .pipette import Pipette
from .types import (HardwareAPILike, CriticalPoint,
                    NoTipAttachedError, TipAttachedError,
                    ExecutionState, ExecutionCancelledError)
from .constants import DROP_TIP_RELEASE_DISTANCE
from .thread_manager import ThreadManager
from .execution_manager import ExecutionManager
from .threaded_async_lock import ThreadedAsyncLock, ThreadedAsyncForbidden

__all__ = [
    'API', 'Controller', 'Simulator', 'Pipette', 'PauseManager',
    'SynchronousAdapter', 'HardwareAPILike', 'CriticalPoint',
    'NoTipAttachedError', 'TipAttachedError', 'DROP_TIP_RELEASE_DISTANCE',
    'ThreadManager', 'ExecutionManager', 'ExecutionState',
    'ExecutionCancelledError', 'ThreadedAsyncLock', 'ThreadedAsyncForbidden'
]



"""Deck data resource provider."""
from dataclasses import dataclass
from typing import List
from typing_extensions import final

from opentrons_shared_data.deck import load as load_deck
from opentrons_shared_data.deck.dev_types import DeckDefinitionV2
from opentrons.protocols.models import LabwareDefinition
from opentrons.types import DeckSlotName
from opentrons.protocols.api_support.constants import STANDARD_DECK, SHORT_TRASH_DECK

from ..types import DeckSlotLocation
from .labware_data_provider import LabwareDataProvider


@final
@dataclass(frozen=True)
class DeckFixedLabware:
    """A labware fixture that is always present on a deck."""

    labware_id: str
    location: DeckSlotLocation
    definition: LabwareDefinition


class DeckDataProvider:
    """Provider class to wrap deck definition and data retrieval."""

    _labware_data: LabwareDataProvider

    def __init__(self, labware_data: LabwareDataProvider) -> None:
        """Initialize a DeckDataProvider."""
        self._labware_data = labware_data

    # NOTE(mc, 2020-11-18): async to allow file reading and parsing to be
    # async on a worker thread in the future
    async def get_deck_definition(
        self,
        *,
        short_fixed_trash: bool = False,
    ) -> DeckDefinitionV2:
        """Get a labware definition given the labware's identification."""
        deck_load_name = SHORT_TRASH_DECK if short_fixed_trash else STANDARD_DECK
        return load_deck(deck_load_name, 2)

    async def get_deck_fixed_labware(
        self,
        deck_definition: DeckDefinitionV2
    ) -> List[DeckFixedLabware]:
        """Get a list of all labware fixtures from a given deck definition."""
        labware = []

        for fixture in deck_definition["locations"]["fixtures"]:
            labware_id = fixture["id"]
            load_name = fixture.get("labware")  # type: ignore[misc]
            slot = fixture.get("slot")  # type: ignore[misc]

            if load_name is not None and slot is not None:
                location = DeckSlotLocation(slot=DeckSlotName.from_primitive(slot))
                definition = await self._labware_data.get_labware_definition(
                    load_name=load_name,
                    namespace="opentrons",
                    version=1,
                )

                labware.append(
                    DeckFixedLabware(
                        labware_id=labware_id,
                        definition=definition,
                        location=location,
                    )
                )

        return labware



import asyncio
import inspect
import logging
import traceback
import sys
from typing import Any, Dict

from opentrons.drivers.smoothie_drivers.driver_3_0 import SmoothieAlarm
from opentrons.protocol_api.contexts import ProtocolContext
from opentrons.protocols.execution.errors import ExceptionInProtocolError
from opentrons.protocols.types import PythonProtocol, MalformedProtocolError
from opentrons.hardware_control import ExecutionCancelledError

MODULE_LOG = logging.getLogger(__name__)


def _runfunc_ok(run_func: Any):
    if not callable(run_func):
        raise SyntaxError("No function 'run(ctx)' defined")
    sig = inspect.Signature.from_callable(run_func)
    if not sig.parameters:
        raise SyntaxError("Function 'run()' does not take any parameters")
    if len(sig.parameters) > 1:
        for name, param in list(sig.parameters.items())[1:]:
            if param.default == inspect.Parameter.empty:
                raise SyntaxError(
                    "Function 'run{}' must be called with more than one "
                    "argument but would be called as 'run(ctx)'"
                    .format(str(sig)))


def _find_protocol_error(tb, proto_name):
    """Return the FrameInfo for the lowest frame in the traceback from the
    protocol.
    """
    tb_info = traceback.extract_tb(tb)
    for frame in reversed(tb_info):
        if frame.filename == proto_name:
            return frame
    else:
        raise KeyError


def run_python(
        proto: PythonProtocol, context: ProtocolContext):
    new_globs: Dict[Any, Any] = {}
    exec(proto.contents, new_globs)
    # If the protocol is written correctly, it will have defined a function
    # like run(context: ProtocolContext). If so, that function is now in the
    # current scope.
    if proto.filename and proto.filename.endswith('zip'):
        filename = 'protocol.ot2.py'
    else:
        filename = proto.filename or '<protocol>'
    try:
        _runfunc_ok(new_globs.get('run'))
    except SyntaxError as se:
        raise MalformedProtocolError(str(se))
    new_globs['__context'] = context
    try:
        exec('run(__context)', new_globs)
    except (SmoothieAlarm, asyncio.CancelledError, ExecutionCancelledError):
        # this is a protocol cancel and shouldn't have special logging
        raise
    except Exception as e:
        exc_type, exc_value, tb = sys.exc_info()
        try:
            frame = _find_protocol_error(tb, filename)
        except KeyError:
            # No pretty names, just raise it
            raise e
        raise ExceptionInProtocolError(e, tb, str(e), frame.lineno)



"""Pipetting command handling."""
from opentrons.hardware_control.api import API as HardwareAPI

from ..state import StateView
from ..types import DeckLocation, WellLocation, WellOrigin
from .movement import MovementHandler


class PipettingHandler:
    """Implementation logic for liquid handling commands."""

    _state: StateView
    _hardware: HardwareAPI
    _movement_handler: MovementHandler

    def __init__(
        self,
        state: StateView,
        hardware: HardwareAPI,
        movement_handler: MovementHandler,
    ) -> None:
        """Initialize a PipettingHandler instance."""
        self._state = state
        self._hardware = hardware
        self._movement_handler = movement_handler

    async def pick_up_tip(
        self,
        pipette_id: str,
        labware_id: str,
        well_name: str,
    ) -> None:
        """Pick up a tip at the specified "well"."""
        # get mount and config data from state and hardware controller
        hw_pipette = self._state.pipettes.get_hardware_pipette(
            pipette_id=pipette_id,
            attached_pipettes=self._hardware.attached_instruments
        )

        # use config data to get tip geometry (length, diameter, volume)
        tip_geometry = self._state.geometry.get_tip_geometry(
            labware_id=labware_id,
            well_name=well_name,
            pipette_config=hw_pipette.config,
        )

        # move the pipette to the top of the tip
        await self._movement_handler.move_to_well(
            pipette_id=pipette_id,
            labware_id=labware_id,
            well_name=well_name,
        )

        # perform the tip pickup routine
        await self._hardware.pick_up_tip(
            mount=hw_pipette.mount,
            tip_length=tip_geometry.effective_length,
            # TODO(mc, 2020-11-12): include these parameters in the request
            presses=None,
            increment=None
        )

        # after a successful pickup, update the hardware controller state
        self._hardware.set_current_tiprack_diameter(
            mount=hw_pipette.mount,
            tiprack_diameter=tip_geometry.diameter
        )
        self._hardware.set_working_volume(
            mount=hw_pipette.mount,
            tip_volume=tip_geometry.volume
        )

    async def drop_tip(
        self,
        pipette_id: str,
        labware_id: str,
        well_name: str,
    ) -> None:
        """Drop a tip at the specified "well"."""
        # get mount and config data from state and hardware controller
        hw_pipette = self._state.pipettes.get_hardware_pipette(
            pipette_id=pipette_id,
            attached_pipettes=self._hardware.attached_instruments
        )

        # get the tip drop location
        well_location = self._state.geometry.get_tip_drop_location(
            labware_id=labware_id,
            pipette_config=hw_pipette.config,
        )

        # move the pipette to tip drop location
        await self._movement_handler.move_to_well(
            pipette_id=pipette_id,
            labware_id=labware_id,
            well_name=well_name,
            well_location=well_location,
        )

        # perform the tip drop routine
        await self._hardware.drop_tip(
            mount=hw_pipette.mount,
            # TODO(mc, 2020-11-12): include this parameter in the request
            home_after=True
        )

    async def aspirate(
        self,
        pipette_id: str,
        labware_id: str,
        well_name: str,
        well_location: WellLocation,
        volume: float,
    ) -> float:
        """Aspirate liquid from a well."""
        # get mount and config data from state and hardware controller
        hw_pipette = self._state.pipettes.get_hardware_pipette(
            pipette_id=pipette_id,
            attached_pipettes=self._hardware.attached_instruments
        )

        ready_to_aspirate = self._state.pipettes.get_is_ready_to_aspirate(
            pipette_id=pipette_id,
            pipette_config=hw_pipette.config,
        )

        current_location = None

        if not ready_to_aspirate:
            await self._movement_handler.move_to_well(
                pipette_id=pipette_id,
                labware_id=labware_id,
                well_name=well_name,
                well_location=WellLocation(origin=WellOrigin.TOP, offset=(0, 0, 0)),
            )

            await self._hardware.prepare_for_aspirate(mount=hw_pipette.mount)

            # set our current deck location to the well now that we've made
            # an intermediate move for the "prepare for aspirate" step
            current_location = DeckLocation(
                pipette_id=pipette_id,
                labware_id=labware_id,
                well_name=well_name)

        await self._movement_handler.move_to_well(
            pipette_id=pipette_id,
            labware_id=labware_id,
            well_name=well_name,
            well_location=well_location,
            current_location=current_location,
        )

        await self._hardware.aspirate(mount=hw_pipette.mount, volume=volume)

        return volume

    async def dispense(
        self,
        pipette_id: str,
        labware_id: str,
        well_name: str,
        well_location: WellLocation,
        volume: float,
    ) -> float:
        """Dispense liquid to a well."""
        hw_pipette = self._state.pipettes.get_hardware_pipette(
            pipette_id=pipette_id,
            attached_pipettes=self._hardware.attached_instruments
        )

        await self._movement_handler.move_to_well(
            pipette_id=pipette_id,
            labware_id=labware_id,
            well_name=well_name,
            well_location=well_location,
        )

        await self._hardware.dispense(mount=hw_pipette.mount, volume=volume)

        return volume



from typing import List

from opentrons.drivers import utils
from opentrons.hardware_control.modules import ThermocyclerStep

from . import types as command_types


def magdeck_engage() -> command_types.MagdeckEngageCommand:
    text = "Engaging Magnetic Module"
    return {
        'name': command_types.MAGDECK_ENGAGE,
        'payload': {'text': text}
    }


def magdeck_disengage() -> command_types.MagdeckDisengageCommand:
    text = "Disengaging Magnetic Module"
    return {
        'name': command_types.MAGDECK_DISENGAGE,
        'payload': {'text': text}
    }


def magdeck_calibrate() -> command_types.MagdeckCalibrateCommand:
    text = "Calibrating Magnetic Module"
    return {
        'name': command_types.MAGDECK_CALIBRATE,
        'payload': {'text': text}
    }


def tempdeck_set_temp(celsius: float) -> command_types.TempdeckSetTempCommand:
    temp = round(float(celsius),
                 utils.TEMPDECK_GCODE_ROUNDING_PRECISION)
    text = f"Setting Temperature Module temperature " \
           f"to {temp} °C (rounded off to nearest integer)"
    return {
        'name': command_types.TEMPDECK_SET_TEMP,
        'payload': {
            'celsius': celsius,
            'text': text
        }
    }


def tempdeck_await_temp(celsius: float) -> command_types.TempdeckAwaitTempCommand:
    text = "Waiting for Temperature Module to reach temperature " \
           "{temp} °C (rounded off to nearest integer)".format(
            temp=round(float(celsius),
                       utils.TEMPDECK_GCODE_ROUNDING_PRECISION))
    return {
        'name': command_types.TEMPDECK_AWAIT_TEMP,
        'payload': {
            'celsius': celsius,
            'text': text
        }
    }


def tempdeck_deactivate() -> command_types.TempdeckDeactivateCommand:
    text = "Deactivating Temperature Module"
    return {
        'name': command_types.TEMPDECK_DEACTIVATE,
        'payload': {'text': text}
    }


def thermocycler_open() -> command_types.ThermocyclerOpenCommand:
    text = "Opening Thermocycler lid"
    return {
        'name': command_types.THERMOCYCLER_OPEN,
        'payload': {'text': text}
    }


def thermocycler_set_block_temp(
        temperature: float,
        hold_time_seconds: float,
        hold_time_minutes: float) -> command_types.ThermocyclerSetBlockTempCommand:
    temp = round(float(temperature), utils.TC_GCODE_ROUNDING_PRECISION)
    text = f'Setting Thermocycler well block temperature to {temp} °C'
    total_seconds = None
    # TODO: BC 2019-09-05 this time resolving logic is partially duplicated
    # in the thermocycler api class definition, with this command logger
    # implementation, there isn't a great way to avoid this, but it should
    # be consolidated as soon as an alternative to the publisher is settled on.
    if hold_time_seconds or hold_time_minutes:
        given_seconds = hold_time_seconds or 0
        given_minutes = hold_time_minutes or 0
        total_seconds = given_seconds + (given_minutes * 60)

        clean_seconds = total_seconds % 60
        clean_minutes = (total_seconds - clean_seconds) / 60
        text += ' with a hold time of '
        if clean_minutes > 0:
            text += f'{clean_minutes} minutes and '
        text += f'{clean_seconds} seconds'
    return {
        'name': command_types.THERMOCYCLER_SET_BLOCK_TEMP,
        'payload': {
            'temperature': temperature,
            'hold_time': total_seconds,
            'text': text
        }
    }


def thermocycler_execute_profile(
        steps: List[ThermocyclerStep],
        repetitions: int) -> command_types.ThermocyclerExecuteProfileCommand:
    text = f'Thermocycler starting {repetitions} repetitions' \
            ' of cycle composed of the following steps: {steps}'
    return {
        'name': command_types.THERMOCYCLER_EXECUTE_PROFILE,
        'payload': {
            'text': text,
            'steps': steps
        }
    }


def thermocycler_wait_for_hold() -> command_types.ThermocyclerWaitForHoldCommand:
    text = "Waiting for hold time duration"
    return {
        'name': command_types.THERMOCYCLER_WAIT_FOR_HOLD,
        'payload': {'text': text}
    }


def thermocycler_wait_for_temp() -> command_types.ThermocyclerWaitForTempCommand:
    text = "Waiting for Thermocycler to reach target"
    return {
        'name': command_types.THERMOCYCLER_WAIT_FOR_TEMP,
        'payload': {'text': text}
    }


def thermocycler_set_lid_temperature(
        temperature: float) -> command_types.ThermocyclerSetLidTempCommand:
    temp = round(float(temperature), utils.TC_GCODE_ROUNDING_PRECISION)
    text = f'Setting Thermocycler lid temperature to {temp} °C'
    return {
        'name': command_types.THERMOCYCLER_SET_LID_TEMP,
        'payload': {'text': text}
    }


def thermocycler_deactivate_lid() -> command_types.ThermocyclerDeactivateLidCommand:
    text = "Deactivating Thermocycler lid heating"
    return {
        'name': command_types.THERMOCYCLER_DEACTIVATE_LID,
        'payload': {'text': text}
    }


def thermocycler_deactivate_block() -> command_types.ThermocyclerDeactivateBlockCommand:
    text = "Deactivating Thermocycler well block heating"
    return {
        'name': command_types.THERMOCYCLER_DEACTIVATE_BLOCK,
        'payload': {'text': text}
    }


def thermocycler_deactivate() -> command_types.ThermocyclerDeactivateCommand:
    text = "Deactivating Thermocycler"
    return {
        'name': command_types.THERMOCYCLER_DEACTIVATE,
        'payload': {'text': text}
    }


def thermocycler_wait_for_lid_temp() -> command_types.ThermocyclerWaitForLidTempCommand:
    text = "Waiting for Thermocycler lid to reach target temperature"
    return {
        'name': command_types.THERMOCYCLER_WAIT_FOR_LID_TEMP,
        'payload': {'text': text}
    }


def thermocycler_close() -> command_types.ThermocyclerCloseCommand:
    text = "Closing Thermocycler lid"
    return {
        'name': command_types.THERMOCYCLER_CLOSE,
        'payload': {'text': text}
    }
